/* This is the single-source-file version of Vlerq for Tcl */

/* %includeDir<../src>% */
/* %includeDir<../src_tcl>% */

/* %include<buffer.c>% */
/*
 * buffer.c - Implementation of a simple temporary buffer.
 */

#include <stdlib.h>
#include <string.h>

/*
 * defs.h - Binding-specific definitions needed by the general code.
 */

typedef struct Tcl_Obj *Object_p;

struct Shared *(GetShared) ();
/*
 * column.h - Definition of sequences, columns, and items.
 */

#include <stdint.h>

#ifdef __sparc__
#define VALUES_MUST_BE_ALIGNED 1
#endif
 
#if WORDS_BIGENDIAN
#define _BIG_ENDIAN 1
#endif

#if DEBUG
#define Assert(x) if (x) ; else FailedAssert(#x, __FILE__, __LINE__);
#else
#define Assert(x)
#endif

#define INVALID_COUNT (-1)

#define PUSH_KEEP_REFS \
  { Buffer keep, *prevkeeps = GetShared()->keeps; \
    InitBuffer(&keep); \
    GetShared()->keeps = &keep;
#define POP_KEEP_REFS \
    ReleaseSequences((Sequence_p*) BufferAsPtr(&keep, 1), \
                     BufferFill(keep) / sizeof (Sequence_p)); \
    ReleaseBuffer(&keep, 0); \
    GetShared()->keeps = prevkeeps; \
  }
    
/* keep in sync with CharAsItemType() in column.c */
typedef enum ItemTypes {
  IT_unknown,
  IT_int,
  IT_wide,
  IT_float,
  IT_double,
  IT_string,
  IT_bytes,
  IT_object,
  IT_column,
  IT_view,
  IT_error
} ItemTypes;

typedef enum ErrorCodes {
  EC_cioor,
  EC_cizwv,
  EC_imvd, 
  EC_nmcw,
  EC_rambe,
  EC_rioor
} ErrorCodes;

typedef struct Shared {
  struct SharedInfo  *info;      /* ext_tcl.c */
  int                 refs;      /* ext_tcl.c */
  struct Buffer      *keeps;     /* column.c */
  long                position;  /* emit.c: current file emit offset */
  struct Buffer      *itembuf;   /* emit.c: current item buffer */
  struct Buffer      *colbuf;    /* emit.c: current column buffer */
  struct Sequence    *empty;     /* view.c */
} Shared, *Shared_p;

typedef struct Sequence *Sequence_p;
typedef struct Column   *Column_p;
typedef union Item      *Item_p;
typedef Sequence_p       View_p;

typedef int       (*Sizer) (Sequence_p seq);
typedef ItemTypes (*Getter) (int row, Item_p item);
typedef void      (*Cleaner) (Sequence_p seq);

typedef struct Sequence {
  int      refcount;  /* reference count */
  int      count;     /* number of items in this sequence */
  Sizer    sizer;     /* function used to determine the row count */
  Getter   getter;    /* function to access a single item */
  Cleaner  cleaner;   /* called during cleanup */
  void    *dummy;     /* needs to be here, see Assert in InitColumn() */
  union { int i; long l; void *p; } data[4]; /* getter-specific use */
} Sequence;

#define C_Count(column) (column.seq->count)

#define IncSeqRef(seq) AdjustSeqRefs(seq, 1)
#define DecSeqRef(seq) AdjustSeqRefs(seq, -1)

typedef struct Column {
  Sequence_p   seq;    /* the sequence which will handle accesses */
  int          pos;    /* position of this column in the view */
} Column;              
                       
typedef union Item {   
  char         b [8];  /* used for raw access and byte-swapping */
  int          q [2];  /* used for hash calculation of wides and doubles */
  int          i;      /* IT_int */
  int64_t      w;      /* IT_wide */
  float        f;      /* IT_float */
  double       d;      /* IT_double */
  const void  *p;      /* used to convert pointers */
  const char  *s;      /* IT_string */
  struct { const uint8_t *ptr; int len; } u; /* IT_bytes */
  Object_p     o;      /* IT_object */
  Column       c;      /* IT_column */
  View_p       v;      /* IT_view */
  ErrorCodes   e;      /* IT_error */
} Item;

Sequence_p  (AdjustSeqRefs) (Sequence_p seq, int count);
void        (AppendToStringColumn) (const void *s, int bytes, Column_p column);
ItemTypes   (CharAsItemType) (char type);
void        (FailedAssert) (const char *msg, const char *file, int line);
Column      (ForceStringColumn) (Column column);
ItemTypes   (GetItem) (int row, Item_p item);
Item        (GetTypedItem) (int row, Column column, ItemTypes type);
Column      (InitColumn) (int count, Getter proc, int auxbytes);
Sequence_p  (KeepRef) (Sequence_p seq);
Column      (NewCountsColumn) (Column column);
Column      (NewIntVec) (int count, int **dataptr);
Column      (NewIotaColumn) (int count);
Column      (NewSequenceColumn) (ItemTypes type, const Sequence_p *p, int n);
Column      (NewStringColumn) (int maxitems, int maxbytes, int istext);
Column      (OmitColumn) (Column omit, int size);
void        (ReleaseSequences) (Sequence_p *refs, int count);
/*
 * buffer.h - Definition of a simple temporary buffer.
 */

typedef struct Buffer *Buffer_p;
typedef struct Overflow *Overflow_p;

typedef struct Buffer {
  union { char *c; int *i; const void **p; } fill;
  char       *limit;
  Overflow_p  head;
  char       *ofill;
  int         ocount;
  char       *result;
  char        buf [128];
  char        slack [8];
} Buffer;

#define ADD_CHAR_TO_BUF(b,x) \
          { char _c = x; \
            if ((b).fill.c < (b).limit) *(b).fill.c++ = _c; \
              else AddToBuffer(&(b), &_c, sizeof _c); }

#define ADD_INT_TO_BUF(b,x) \
          { int _i = x; \
            if ((b).fill.c < (b).limit) *(b).fill.i++ = _i; \
              else AddToBuffer(&(b), &_i, sizeof _i); }

#define ADD_PTR_TO_BUF(b,x) \
          { const void *_p = x; \
            if ((b).fill.c < (b).limit) *(b).fill.p++ = _p; \
              else AddToBuffer(&(b), &_p, sizeof _p); }

#define BufferFill(b) ((b).ocount + ((b).fill.c - (b).buf))

void   InitBuffer(Buffer_p bp);
void   ReleaseBuffer(Buffer_p bp, int keep);
void   AddToBuffer(Buffer_p bp, const void *data, int len);
char  *BufferAsPtr(Buffer_p bp, int fast);
Column BufferAsIntCol(Buffer_p bp);

typedef struct Overflow {
  Overflow_p p;
  char       b[4096];
} Overflow;

void InitBuffer(Buffer_p bp) {
  bp->fill.c = bp->buf;
  bp->limit = bp->buf + sizeof bp->buf;
  bp->head = 0;
  bp->ofill = 0;
  bp->ocount = 0;
  bp->result = 0;
}

void ReleaseBuffer(Buffer_p bp, int keep) {
  while (bp->head != 0) {
    Overflow_p op = bp->head;
    bp->head = op->p;
    free(op);
  }
  if (!keep && bp->result != 0)
    free(bp->result);
}

void AddToBuffer(Buffer_p bp, const void *data, int len) {
  int n;
  while (len > 0) {
    if (bp->fill.c >= bp->limit) {
      if (bp->head == 0 || bp->ofill >= bp->head->b + sizeof bp->head->b) {
        Overflow_p op = (Overflow_p) malloc(sizeof(Overflow));
        op->p = bp->head;
        bp->head = op;
        bp->ofill = op->b;
      }
      memcpy(bp->ofill, bp->buf, sizeof bp->buf);
      bp->ofill += sizeof bp->buf;
      bp->ocount += sizeof bp->buf;
      n = bp->fill.c - bp->slack;
      memcpy(bp->buf, bp->slack, n);
      bp->fill.c = bp->buf + n;
    }
    n = len;
    if (n > bp->limit - bp->fill.c)
      n = bp->limit - bp->fill.c; /* TODO: copy big chunks to overflow */
    memcpy(bp->fill.c, data, n);
    bp->fill.c += n;
    data = (const char*) data + n;
    len -= n;
  }
}

char *BufferAsPtr(Buffer_p bp, int fast) {
  int n = bp->fill.c - bp->buf, len = bp->ocount + n;
  if (fast && n == len)
    return bp->buf;
  if (bp->result == 0)
    bp->result = malloc(len);
  len -= n;
  memcpy(bp->result + len, bp->buf, n);
  if (bp->head != 0) {
    Overflow_p curr = bp->head;
    n = bp->ofill - curr->b;
    do {
      len -= n;
      memcpy(bp->result + len, curr->b, n);
      curr = curr->p;
      n = sizeof bp->head->b;
    } while (curr != 0);
  }
  return bp->result;
}

Column BufferAsIntCol(Buffer_p bp) {
  int count, *data;
  Column result;
  
  count = BufferFill(*bp) / sizeof (int);
  result = NewIntVec(count, &data);
  memcpy(data, BufferAsPtr(bp, 1), BufferFill(*bp));
  ReleaseBuffer(bp, 0); /* TODO: try to avoid extra alloc + copy */
  
  return result;
}
/* %include<column.c>% */
/*
 * column.c - Implementation of sequences, columns, and items.
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/*
 * view.h - Definition of views.
 */

typedef enum MetaCols {
  MC_name, MC_type, MC_subv, MC_limit
} MetaCols;

#define V_Cols(view) (*(Column_p*) (view)->data)
#define V_Types(view) (*(char**) ((view)->data+1))
#define V_Meta(view) (*(View_p*) ((view)->data+2))

#define IncViewRef(view) AdjustSeqRefs(view, 1)
#define DecViewRef(view) AdjustSeqRefs(view, -1)
#define ViewWidth(view) ViewSize(V_Meta(view))
#define ViewCol(view,col) (V_Cols(view)[col])
#define ViewColType(view,col) ((ItemTypes) V_Types(view)[col])
#define ViewCompat(view1,view2) MetaIsCompatible(V_Meta(view1), V_Meta(view2))

View_p (CloneView) (View_p view);
View_p (ColMapView) (View_p view, Column mapcol);
View_p (ColOmitView) (View_p view, Column mapcol);
View_p (EmptyMetaView) ();
View_p (FirstView) (View_p view, int count);
char   (GetColType) (View_p meta, int col);
View_p (IndirectView) (View_p meta, Sequence_p seq);
View_p (LastView) (View_p view, int count);
View_p (NameColView) (View_p view);
View_p (NewView) (View_p meta);
View_p (NoColumnView) (int rows);
View_p (PairView) (View_p view1, View_p view2);
View_p (TakeView) (View_p view, int count);
int    (MetaIsCompatible) (View_p meta1, View_p meta2);
int    (ViewCompare) (View_p view1, View_p view2);
int    (ViewSize) (View_p view);
void   (ViewStructure) (View_p meta, Buffer_p buffer);
void   (ReleaseUnusedViews) ();
void   (SetViewColumns) (View_p view, int first, int count, Column src);
View_p (TagView) (View_p view, const char* name);
/*
 * getters.h - Definition of several simple getter functions.
 */

Getter (PickIntGetter) (int bits);
Getter (PickFixedWidthGetter) (int bytes, int rows, int real, int flip);

Sequence_p AdjustSeqRefs (Sequence_p seq, int count) {
  if (seq != NULL) {
    seq->refcount += count;
    if (seq->refcount <= 0) {
      if (seq->cleaner != NULL)
        seq->cleaner(seq);
      free(seq);
      seq = NULL;
    }
  }
  return seq;
}

Sequence_p KeepRef (Sequence_p seq) {
  Shared_p sh = GetShared();
  
  ADD_PTR_TO_BUF(*sh->keeps, seq);
  return AdjustSeqRefs(seq, 1);
}

static int DefaultSizer (Sequence_p seq) {
  return seq->count;
}

Column InitColumn (int count, Getter proc, int auxbytes) {
  Column column;
  
  column.seq = KeepRef((Sequence_p) calloc(1, sizeof(Sequence) + auxbytes));
  column.seq->count = count;
  column.seq->sizer = DefaultSizer;
  column.seq->getter = proc;

  /* default for data[0] is to point to the aux bytes after the sequence */
  column.seq->data[0].p = column.seq + 1;
  
  /* make sure 64-bit values have a "native" 8-byte alignment */
  Assert(((int) column.seq->data[0].p) % 8 == 0);

  column.pos = INVALID_COUNT;
  return column;
}

void ReleaseSequences (Sequence_p *refs, int count) {
  while (--count >= 0)
    AdjustSeqRefs(refs[count], -1);
}

Column NewIntVec (int count, int **dataptr) {
  Column result = InitColumn(count, PickIntGetter(32), 4 * count);
  if (dataptr != NULL)
    *dataptr = (int*) result.seq->data[0].p;
  return result;
}

ItemTypes GetItem (int row, Item_p item) {
  return item->c.seq->getter(row, item);
}

Item GetTypedItem (int row, Column column, ItemTypes type) {
  Item item;
  ItemTypes got;
  
  item.c = column;
  got = GetItem(row, &item);
  Assert(got == type);
    
  return item;
}

#if 0 /* unused */
const char *ItemTypeAsString (ItemTypes type) {
  static const char *typeTable[] = {
    "",  /* IT_unknown */
    "I", /* IT_int */
    "L", /* IT_wide */
    "D", /* IT_double */
    "S", /* IT_string */
    "O", /* IT_object */
    "C", /* IT_column */
    "V", /* IT_view */
    "E", /* IT_error */
  };

  return typeTable[type];
}
#endif

ItemTypes CharAsItemType (char type) {
  switch (type) {
    case 0:   return IT_unknown;
    case 'I': return IT_int;
    case 'L': return IT_wide;
    case 'F': return IT_float;
    case 'D': return IT_double;
    case 'S': return IT_string;
    case 'B': return IT_bytes;
    case 'O': return IT_object;
    case 'C': return IT_column;
    case 'V': return IT_view;
  }
  return IT_error;
}

static void StringCleaner (Sequence_p seq) {
  DecSeqRef((Sequence_p) seq->data[2].p);
}

static ItemTypes StringGetter (int row, Item_p item) {
  char **ptrs = (char**) item->c.seq->data[0].p;
  item->s = ptrs[row];
  return IT_string;
}

static ItemTypes BytesGetter (int row, Item_p item) {
  char **ptrs = (char**) item->c.seq->data[0].p;
  item->u.ptr = (const uint8_t*) ptrs[row];
  item->u.len = ptrs[row+1] - ptrs[row];
  return IT_bytes;
}

Column NewStringColumn (int maxitems, int maxbytes, int istext) {
  int ptrbytes;
  char **ptrs;
  Column result;

  ptrbytes = (maxitems + 1) * sizeof(char*);
  result = InitColumn(0, istext ? StringGetter : BytesGetter,
                                                        ptrbytes + maxbytes);
  result.seq->cleaner = StringCleaner;
  /* data[0] is a vector of string pointers */
  /* data[1] is not used */
  /* data[2] is an optional pointer to a hash map sequence, see StringLookup */
  /* data[3] is the optional prime factor used in hash map, see StringLookup */
  result.seq->data[1].i = 0;
  
  /* last pointer points to first unused byte */
  ptrs = (char**) result.seq->data[0].p;
  ptrs[0] = (char*) ptrs + ptrbytes;
  return result;
}

void AppendToStringColumn (const void *string, int bytes, Column_p column) {
  char **ptrs = (char**) column->seq->data[0].p;
  char *bufend = ptrs[column->seq->count];

  if (bytes < 0)
    bytes = strlen((const char*) string) + 1;
  memcpy(bufend, string, bytes);
  ptrs[++column->seq->count] = bufend + bytes;
}

static int PreflightStringLengths (Column column) {
  int i, count = column.seq->count, total = 0;

  for (i = 0; i < count; ++i)
    total +=  strlen(GetTypedItem(i, column, IT_string).s) + 1;

  return total;
}

Column ForceStringColumn (Column column) {
  int r, rows;
  Column result;
  
  /* this code is needed to always make name column of meta views string cols */
  
  if (column.seq == NULL || column.seq->getter == StringGetter)
    return column;
    
  rows = column.seq->count;
  result = NewStringColumn(rows, PreflightStringLengths(column), 1);
  
  for (r = 0; r < rows; ++r)
    AppendToStringColumn(GetTypedItem(r, column, IT_string).s, -1, &result);
  
  return result;
}

static ItemTypes IotaGetter (int row, Item_p item) {
  item->i = row;
  return IT_int;
}

Column NewIotaColumn (int count) {
  return InitColumn(count, IotaGetter, 0);
}

static void SequenceCleaner (Sequence_p seq) {
  int i, count;
  View_p *items;

  count = seq->count;
  items = (View_p*) seq->data[0].p;

  for (i = 0; i < count; ++i)
    DecViewRef(items[i]);
}

static ItemTypes SequenceGetter (int row, Item_p item) {
  void **items = (void**) item->c.seq->data[0].p;
  ItemTypes type = item->c.seq->data[1].i;

  switch (type) {

    case IT_view:
      item->v = items[row];
      return IT_view;

    case IT_column:
      item->c.seq = items[row];
      item->c.pos = INVALID_COUNT;
      return IT_column;

    default:
      return IT_unknown;
  }
}

Column NewSequenceColumn (ItemTypes type, const Sequence_p *items, int count) {
  int i, bytes;
  Column result;

  bytes = count * sizeof(View_p);
  result = InitColumn(count, SequenceGetter, bytes);
  result.seq->cleaner = SequenceCleaner;

  /* data[0] points a list of pointers */
  /* data[1] is the type of the returned items */
  result.seq->data[1].i = type;

  if (items != NULL) {
    memcpy(result.seq->data[0].p, items, bytes);
    for (i = 0; i < count; ++i)
      IncViewRef(items[i]);
  }

  return result;
}

static ItemTypes CountsGetter (int row, Item_p item) {
  const View_p *items = (const View_p*) item->c.seq->data[0].p;

  item->i = ViewSize(items[row]);
  return IT_int;
}

Column NewCountsColumn (Column column) {
  int r, rows;
  View_p *items;
  Column result;
  
  rows = column.seq->count;  
  result = NewSequenceColumn(IT_view, NULL, rows);
  result.seq->getter = CountsGetter;

  items = result.seq->data[0].p;
  for (r = 0; r < rows; ++r)
    items[r] = IncViewRef(GetTypedItem(r, column, IT_view).v);

  return result;
}

Column OmitColumn (Column omit, int size) {
  int i, *data, next = 0, omitcount = omit.seq->count, limit;
  Column result;
  
  result = NewIntVec(size - omitcount, &data);
  for (i = 0; i < omitcount; ++i) {
    limit = GetTypedItem(i, omit, IT_int).i;
    while (next < limit) 
      *data++ = next++;
    ++next;
  }
  while (next < size) 
    *data++ = next++;
    
  return result;
}
/* %include<emit.c>% */
/*
 * emit.c - Implementation of file output commands.
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/*
 * emit.h - Definition of file output commands.
 */

typedef enum EmitTypes {
  ET_mem
} EmitTypes;

typedef struct EmitItem {
  EmitTypes type;
  int size;
  const void *data;
} EmitItem, *EmitItem_p;

long (EmitFull) (View_p view, Buffer_p buffer);
void (EmitView) (View_p view, int describe);
void (MetaAsDesc) (View_p meta, Buffer_p buffer);

static void EmitBlock (const void* data, int size) {
  if (size > 0) {
    EmitItem item;
    Shared_p sh = GetShared();

    item.type = ET_mem;
    item.size = size;
    item.data = data;
    AddToBuffer(sh->itembuf, &item, sizeof item);

    sh->position += size;
  }
}

static void *EmitCopy (const void* data, int size) {
  void *buf = NULL;
  if (size > 0) {
    buf = memcpy(malloc(size), data, size);
    EmitBlock(buf, size);
  }
  return buf;
}

void MetaAsDesc (View_p meta, Buffer_p buffer) {
  int r, rows = ViewSize(meta);
  Column names, types, subvs;
  char type;
  const char *name;
  View_p subv;

  names = ViewCol(meta, MC_name);
  types = ViewCol(meta, MC_type);
  subvs = ViewCol(meta, MC_subv);

  for (r = 0; r < rows; ++r) {
    if (r > 0)
      ADD_CHAR_TO_BUF(*buffer, ',');

    name = GetTypedItem(r, names, IT_string).s;
    type = *GetTypedItem(r, types, IT_string).s;
    subv = GetTypedItem(r, subvs, IT_view).v;

    AddToBuffer(buffer, name, strlen(name));
    if (type == 'V' && ViewSize(subv) > 0) {
      ADD_CHAR_TO_BUF(*buffer, '[');
      MetaAsDesc(subv, buffer);
      ADD_CHAR_TO_BUF(*buffer, ']');
    } else {
      ADD_CHAR_TO_BUF(*buffer, ':');
      ADD_CHAR_TO_BUF(*buffer, type);
    }
  }
}

static void EmitVarInt (long value) {
  int n;
  Shared_p sh = GetShared();

  if (value < 0) {
    ADD_CHAR_TO_BUF(*sh->colbuf, 0);
    value = ~value;
  }

  for (n = 7; (value >> n) > 0; n += 7)
    ;
  while (n > 7) {
    n -= 7;
    ADD_CHAR_TO_BUF(*sh->colbuf, (value >> n) & 0x7F);
  }
  ADD_CHAR_TO_BUF(*sh->colbuf, (value & 0x7F) | 0x80);
}

static void EmitPair (long offset) {
  long size;
  Shared_p sh = GetShared();

  size = sh->position - offset;
  EmitVarInt(size);
  if (size > 0)
    EmitVarInt(offset);
}

static int TopBit (unsigned v) {
#define Vn(x) (v < (1<<x))
  return Vn(16) ? Vn( 8) ? Vn( 4) ? Vn( 2) ? Vn( 1) ? (int)v-1 : 1
                                           : Vn( 3) ?  2 :  3
                                  : Vn( 6) ? Vn( 5) ?  4 :  5
                                           : Vn( 7) ?  6 :  7
                         : Vn(12) ? Vn(10) ? Vn( 9) ?  8 :  9
                                           : Vn(11) ? 10 : 11
                                  : Vn(14) ? Vn(13) ? 12 : 13
                                           : Vn(15) ? 14 : 15
                : Vn(24) ? Vn(20) ? Vn(18) ? Vn(17) ? 16 : 17
                                           : Vn(19) ? 18 : 19
                                  : Vn(22) ? Vn(21) ? 20 : 21
                                           : Vn(23) ? 22 : 23
                         : Vn(28) ? Vn(26) ? Vn(25) ? 24 : 25
                                           : Vn(27) ? 26 : 27
                                  : Vn(30) ? Vn(29) ? 28 : 29
                                           : Vn(31) ? 30 : 31;
#undef Vn
}

static int MinWidth (int lo, int hi) {
  if (lo > 0) lo = 0;
  if (hi < 0) hi = 0;
  lo = "04444444455555555666666666666666"[TopBit(~lo)+2] & 7;
  hi = "01233444555555556666666666666666"[TopBit(hi)+1] & 7;
  return lo > hi ? lo : hi;
}

static int FixFudge (int n, int w) {
  static char fudges[3][4] = {  /* n:  1:  2:  3:  4: */
    {3,3,4,5},      /* 1-bit entries:  1b  2b  3b  4b */
    {5,5,1,1},      /* 2-bit entries:  2b  4b  6b  8b */
    {6,1,2,2},      /* 4-bit entries:  4b  8b 12b 16b */
  };
  return n <= 4 && w <= 3 ? fudges[w-1][n-1] : ((n<<w)+14)>>4;
}

static int *PackedIntVec (int *data, int rows, int *outsize) {
  int r, width, bytes, lo = 0, hi = 0, bits, *limit, *result;

  for (r = 0; r < rows; ++r) {
    if (data[r] < lo) lo = data[r];
    if (data[r] > hi) hi = data[r];
  }

  width = MinWidth(lo, hi);

  if (width >= 6) {
    bytes = rows * (1 << (width-4));
    result = malloc(bytes);
    memcpy(result, data, bytes);
  } else if (rows > 0 && width > 0) {
    bytes = FixFudge(rows, width);
    result = malloc(bytes);
    memset(result, 0, bytes);

    limit = data + rows;
    bits = 0;

    switch (width) {

      case 1: { /* 1 bit, 8 per byte */
        char* q = (char*) result;
        while (data < limit) {
          *q |= (*data++ & 1) << bits; ++bits; q += bits >> 3; bits &= 7;
        }
        break;
      }

      case 2: { /* 2 bits, 4 per byte */
        char* q = (char*) result;
        while (data < limit) {
          *q |= (*data++ & 3) << bits; bits += 2; q += bits >> 3; bits &= 7;
        }
        break;
      }

      case 3: { /* 4 bits, 2 per byte */
        char* q = (char*) result;
        while (data < limit) {
          *q |= (*data++ & 15) << bits; bits += 4; q += bits >> 3; bits &= 7;
        }
        break;
      }

      case 4: { /* 1-byte (char) */
        char* q = (char*) result;
        while (data < limit)
          *q++ = (char) *data++;
        break;
      }

      case 5: { /* 2-byte (short) */
        short* q = (short*) result;
        while (data < limit)
          *q++ = (short) *data++;
        break;
      }
    }
  } else {
    bytes = 0;
    result = NULL;
  }

  *outsize = bytes;
  return result;
}

static void EmitFixCol (Column column, ItemTypes type) {
  int r, rows, *tempvec, bufsize;
  void *buffer;
  long basepos;
  Shared_p sh = GetShared();

  rows = C_Count(column);

  switch (type) {

    case IT_int:
      bufsize = rows * sizeof (int);
      tempvec = malloc(bufsize);
      for (r = 0; r < rows; ++r)
        tempvec[r] = GetTypedItem(r, column, type).i;
      buffer = PackedIntVec(tempvec, rows, &bufsize);
      free((char*) tempvec); /* TODO: avoid extra copy & malloc */
      break;

    case IT_wide:
      bufsize = rows * sizeof (int64_t);
      buffer = malloc(bufsize);
      for (r = 0; r < rows; ++r)
        ((int64_t*) buffer)[r] = GetTypedItem(r, column, type).w;
      break;

    case IT_float:
      bufsize = rows * sizeof (float);
      buffer = malloc(bufsize);
      for (r = 0; r < rows; ++r)
        ((float*) buffer)[r] = GetTypedItem(r, column, type).f;
      break;

    case IT_double:
      bufsize = rows * sizeof (double);
      buffer = malloc(bufsize);
      for (r = 0; r < rows; ++r)
        ((double*) buffer)[r] = GetTypedItem(r, column, type).d;
      break;

    default: Assert(0); return;
  }

  /* start using 8-byte alignment once the emitted data reaches 1 Mb */
  /* only do this for vectors >= 64 bytes, so worst-case waste is under 10% */
  
  Assert(!(bufsize > 0 && rows == 0));
  if (bufsize >= 64 && bufsize / rows >= 2 && sh->position >= 1024 * 1024) 
    EmitCopy("\0\0\0\0\0\0\0", 7 & - sh->position);
    
  basepos = sh->position;
  EmitBlock(buffer, bufsize);
  EmitPair(basepos);
}

static void EmitVarCol (Column column, int istext) {
  int r, rows, bytes, *sizevec;
  long basepos;
  Item item;
  Buffer buffer;
  Column sizes;
  Shared_p sh = GetShared();

  InitBuffer(&buffer);
  rows = C_Count(column);
  sizes = NewIntVec(rows, &sizevec);

  if (istext)
    for (r = 0; r < rows; ++r) {
      item = GetTypedItem(r, column, IT_string);
      bytes = strlen(item.s);
      if (bytes > 0)
        AddToBuffer(&buffer, item.s, ++bytes);
      sizevec[r] = bytes;
    }
  else
    for (r = 0; r < rows; ++r) {
      item = GetTypedItem(r, column, IT_bytes);
      AddToBuffer(&buffer, item.u.ptr, item.u.len);
      sizevec[r] = item.u.len;
    }

  basepos = sh->position;
  EmitBlock(BufferAsPtr(&buffer, 0), BufferFill(buffer));
  EmitPair(basepos);
  
  if (BufferFill(buffer) > 0)
    EmitFixCol(sizes, 1);

  ReleaseBuffer(&buffer, 1);
  EmitVarInt(0); /* no memos */
}

static void EmitSubCol (Column column, int describe) {
  int r, rows;
  long basepos;
  View_p view;
  Buffer newcolbuf;
  Buffer_p origcolbuf;
  Shared_p sh = GetShared();

  origcolbuf = sh->colbuf;
  sh->colbuf = &newcolbuf;
  InitBuffer(sh->colbuf);

  rows = C_Count(column);

  for (r = 0; r < rows; ++r) {
    view = GetTypedItem(r, column, IT_view).v;
    EmitView(view, describe);
  }

  sh->colbuf = origcolbuf;

  basepos = sh->position;
  EmitBlock(BufferAsPtr(&newcolbuf, 0), BufferFill(newcolbuf));
  EmitPair(basepos);

  ReleaseBuffer(&newcolbuf, 1);
}

void EmitView (View_p view, int describe) {
  int rows;
  Shared_p sh = GetShared();

  EmitVarInt(0);

  if (describe) {
    Buffer desc;

    InitBuffer(&desc);
    MetaAsDesc(V_Meta(view), &desc);

    EmitVarInt(BufferFill(desc));
    AddToBuffer(sh->colbuf, BufferAsPtr(&desc, 1), BufferFill(desc));

    ReleaseBuffer(&desc, 0);
  }

  rows = ViewSize(view);
  EmitVarInt(rows);

  if (rows > 0) {
    int c, cols = ViewWidth(view);
    ItemTypes type;
    Column column;
    View_p subv, meta = V_Meta(view);

    for (c = 0; c < cols; ++c) {
      column = ViewCol(view,c);
      type = ViewColType(view, c);
      switch (type) {
        
        case IT_int:
        case IT_wide:
        case IT_float:
        case IT_double:
          EmitFixCol(column, type); 
          break;
          
        case IT_string:
          EmitVarCol(column, 1);
          break;
          
        case IT_bytes:
          EmitVarCol(column, 0);
          break;
          
        case IT_view:
          subv = GetTypedItem(c, ViewCol(meta, MC_subv), IT_view).v;
          EmitSubCol(column, ViewSize(subv) == 0);
          break;
          
        default: Assert(0);
      }
    }
  }
}

static void SetBigEndian32 (char* dest, int value) {
  dest[0] = (char) (value >> 24);
  dest[1] = (char) (value >> 16);
  dest[2] = (char) (value >> 8);
  dest[3] = (char) value;
}

long EmitFull (View_p view, Buffer_p buffer) {
  long rootpos, tailpos;
  char tail[16], *fixup;
  struct Head { short a; char b; char c; char d[4]; } head;
  Buffer newcolbuf;
  Shared_p sh = GetShared();

  sh->position = 0;
  sh->itembuf = buffer;

  head.a = 'J' + ('L' << 8);
  head.b = 0x1A;
  head.c = 0;
  fixup = ((struct Head*) EmitCopy(&head, sizeof head))->d;

  sh->colbuf = &newcolbuf;
  InitBuffer(sh->colbuf);

  EmitView(view, 1);

  rootpos = sh->position;
  EmitCopy(BufferAsPtr(sh->colbuf, 0), BufferFill(*sh->colbuf));

  ReleaseBuffer(sh->colbuf, 1);

  sh->colbuf = NULL;

  tailpos = sh->position;

  SetBigEndian32(tail, 0x80000000);
  SetBigEndian32(tail+4, tailpos);
  SetBigEndian32(tail+8, 0x80000000 + tailpos - rootpos);
  SetBigEndian32(tail+12, rootpos);
  EmitCopy(tail, sizeof tail);

  SetBigEndian32(fixup, tailpos + sizeof tail);
  return tailpos + sizeof tail;
}
/* %include<file.c>% */
/*
 * file.c - Implementation of memory-mapped file access.
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#else
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif

/*
 * file.h - Definition of memory-mapped file access.
 */

typedef Sequence_p MappedFile_p;

#define MF_Data(map) ((const char*) ((map)->data[0].p))
#define MF_Length(map) ((map)->data[1].l)

MappedFile_p (InitMappedFile) (const char *data, long length, Cleaner fun);
View_p       (MappedToView) (MappedFile_p map);
View_p       (MapSubview) (MappedFile_p map, int offset, View_p meta);
View_p       (OpenDataFile) (const char *filename);
MappedFile_p (OpenMappedFile) (const char *filename);

MappedFile_p InitMappedFile (const char *data, long length, Cleaner fun) {
  MappedFile_p map;
  
  /* data[0] points to the current start of the mapped area */
  /* data[1] is the current length of the mapped area */
  /* data[2] points to the original mapped area */
  /* data[3] is available for storing an Object_p, see ext_tcl.c */
  map = InitColumn(0, NULL, 0).seq;
  map->data[0].p = (void*) data;
  map->data[1].l = length;
  map->data[2].p = (void*) data;
  map->cleaner = fun;
  return map;
}

static void MappedFileCleaner (MappedFile_p map) {
#if WIN32
  UnmapViewOfFile(map->data[2].p);
#else
  int offset = (char*) map->data[0].p - (char*) map->data[2].p;
  munmap(map->data[2].p, (long) map->data[1].l + offset);
#endif
}

MappedFile_p OpenMappedFile (const char *filename) {
  const char *data = NULL;
  int length = -1;
    
#if WIN32
  {
    DWORD n;
    HANDLE h, f = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, 0,
                              OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    if (f != INVALID_HANDLE_VALUE) {
      h = CreateFileMapping(f, 0, PAGE_READONLY, 0, 0, 0);
      if (h != INVALID_HANDLE_VALUE) {
        n = GetFileSize(f, 0);
        data = MapViewOfFile(h, FILE_MAP_READ, 0, 0, n);
        if (data != NULL)
          length = n;
        CloseHandle(h);
      }
      CloseHandle(f);
    }
  }
#else
  {
    struct stat sb;
    int fd = open(filename, O_RDONLY);
    if (fd != -1) {
      if (fstat(fd, &sb) == 0) {
        data = mmap(0, sb.st_size, PROT_READ, MAP_SHARED, fd, 0);
        if (data != MAP_FAILED)
          length = sb.st_size;
      }
      close(fd);
    }
  }
#endif

  if (length < 0)
    return NULL;
    
  return InitMappedFile(data, length, MappedFileCleaner);
}

static void AdjustMappedFile (MappedFile_p map, int offset) {
  map->data[0].p = (void*) (MF_Data(map) + offset);
  map->data[1].p = (void*) (MF_Length(map) - offset);
}

static int IsReversedEndian(MappedFile_p map) {
#if _BIG_ENDIAN
  return *MF_Data(map) == 'J';
#else
  return *MF_Data(map) == 'L';
#endif
}

static long GetVarInt (const char **cp) {
  int8_t b;
  long v = 0;
  do {
    b = *(*cp)++;
    v = (v << 7) + b;
  } while (b >= 0);
  return v + 128;
}

static long GetVarPair(const char** cp) {
  long n = GetVarInt(cp);
  if (n > 0 && GetVarInt(cp) == 0)
    *cp += n;
  return n;
}

static View_p DescAsMeta (const char** desc, const char* end) {
  int c, len, cols = 0, namebytes = 0;
  char sep, *buf, *ptr, **strings;
  Sequence_p *views;
  Buffer names, types, subvs;
  Column tempcol;
  View_p result, empty;

  InitBuffer(&names);
  InitBuffer(&types);
  InitBuffer(&subvs);
  
  len = end - *desc;
  buf = memcpy(malloc(len+1), *desc, len);
  buf[len] = ',';
  
  empty = EmptyMetaView();
  result = NewView(V_Meta(empty));
  
  for (ptr = buf; ptr < buf + len; ++ptr) {
    const char *s = ptr;
    while (strchr(":,[]", *ptr) == 0)
      ++ptr;
    sep = *ptr;
    *ptr = 0;

    ADD_PTR_TO_BUF(names, s);
    namebytes += ptr - s + 1;
     
    switch (sep) {

      case '[':
        ADD_PTR_TO_BUF(types, "V");
        ++ptr;
        ADD_PTR_TO_BUF(subvs, DescAsMeta((const char**) &ptr, buf + len));
        sep = *++ptr;
        break;

      case ':':
        ptr += 2;
        sep = *ptr;
        *ptr = 0;
        ADD_PTR_TO_BUF(types, ptr - 1);
        ADD_PTR_TO_BUF(subvs, empty);
        break;

      default:
        ADD_PTR_TO_BUF(types, "S");
        ADD_PTR_TO_BUF(subvs, empty);
        break;
    }
    
    ++cols;
    if (sep != ',')
      break;
  }

  strings = (char**) BufferAsPtr(&names, 1);
  tempcol = NewStringColumn(cols, namebytes, 1);
  for (c = 0; c < cols; ++c)
    AppendToStringColumn(strings[c], -1, &tempcol);
  SetViewColumns(result, MC_name, 1, tempcol);
  ReleaseBuffer(&names, 0);
  
  strings = (char**) BufferAsPtr(&types, 1);
  tempcol = NewStringColumn(cols, 2 * cols, 1);
  for (c = 0; c < cols; ++c)
    AppendToStringColumn(strings[c], -1, &tempcol);
  SetViewColumns(result, MC_type, 1, tempcol);
  ReleaseBuffer(&types, 0);
  
  views = (Sequence_p*) BufferAsPtr(&subvs, 1);
  tempcol = NewSequenceColumn(IT_view, views, cols);
  SetViewColumns(result, MC_subv, 1, tempcol);
  ReleaseBuffer(&subvs, 0);
  
  free(buf);
  *desc += ptr - buf;
  return result;
}

static void MappedViewCleaner (Sequence_p seq) {
  int i, count;
  const View_p *subviews = seq->data[0].p;
  
  count = ((Sequence_p) seq->data[1].p)->count;
  for (i = 0; i < count; ++i)
    DecViewRef(subviews[i]);
  
  DecSeqRef((Sequence_p) seq->data[1].p);
  DecSeqRef((MappedFile_p) seq->data[2].p);
  DecViewRef((View_p) seq->data[3].p);
}

static ItemTypes MappedViewGetter (int row, Item_p item) {
  View_p *subviews = item->c.seq->data[0].p;
  
  if (subviews[row] == NULL) {
    const int *offsets = ((Sequence_p) item->c.seq->data[1].p)->data[0].p;
    MappedFile_p map = item->c.seq->data[2].p;
    View_p meta = item->c.seq->data[3].p;
    subviews[row] = IncViewRef(MapSubview(map, offsets[row], meta));
  }
  
  item->v = subviews[row];
  return IT_view;
}

static Column MappedViewCol (MappedFile_p map, int rows, const char **nextp, View_p meta) {
  int r, c, cols, subcols, *offsets;
  long colsize, colpos;
  const char *next;
  Sequence_p offseq;
  Column result;
  
  offseq = NewIntVec(rows, &offsets).seq;
  cols = ViewSize(meta);
  
  colsize = GetVarInt(nextp);
  colpos = colsize > 0 ? GetVarInt(nextp) : 0;
  next = MF_Data(map) + colpos;
  
  for (r = 0; r < rows; ++r) {
    offsets[r] = next - MF_Data(map);
    GetVarInt(&next);
    if (cols == 0) {
      long desclen = GetVarInt(&next);
      const char *desc = next;
      next += desclen;
      meta = DescAsMeta(&desc, next);
    }
    if (GetVarInt(&next) > 0) {
      subcols = ViewSize(meta);
      for (c = 0; c < subcols; ++c)
        switch (GetColType(meta, c)) {
          case 'B': case 'S':
            if (GetVarPair(&next))
              GetVarPair(&next);
            /* fall through */
          default:
            GetVarPair(&next);
        }
    }
  }
  
  /* data[0] points to a subview cache */
  /* data[1] points to a sequence owning the offest vector */
  /* data[2] points to the mapped file */
  /* data[3] points to the meta view */
  result = InitColumn(rows, MappedViewGetter, rows * sizeof(View_p));
  result.seq->cleaner = MappedViewCleaner;
  result.seq->data[1].p = IncSeqRef(offseq);
  result.seq->data[2].p = IncSeqRef(map);
  result.seq->data[3].p = IncViewRef(cols > 0 ? meta : EmptyMetaView());
  
  return result;
}

static void MappedFixedCleaner (Sequence_p seq) {
  DecSeqRef((MappedFile_p) seq->data[1].p);
}

static Column MappedFixedCol (MappedFile_p map, int rows, const char **nextp, int isreal) {
  long colsize, colpos;
  Getter getter;
  Column result;

  colsize = GetVarInt(nextp);
  colpos = colsize > 0 ? GetVarInt(nextp) : 0;
  getter = PickFixedWidthGetter(colsize, rows, isreal, IsReversedEndian(map));
  
  /* data[0] points to the mapped data */
  /* data[1] points to the mapped file */
  result = InitColumn(rows, getter, 0);
  result.seq->cleaner = MappedFixedCleaner;
  result.seq->data[0].p = (void*) (MF_Data(map) + colpos);
  result.seq->data[1].p = IncSeqRef(map);
  
  return result;
}

static void MappedStringCleaner (Sequence_p seq) {
  DecSeqRef((Sequence_p) seq->data[0].p);
  DecSeqRef((MappedFile_p) seq->data[2].p);
  DecSeqRef((Sequence_p) seq->data[3].p);
}

static ItemTypes MappedStringGetter (int row, Item_p item) {
  const int *offsets = (const int*) item->c.seq->data[1].p;
  MappedFile_p map = (MappedFile_p) item->c.seq->data[2].p;

  if (offsets[row] >= 0)
    item->s = MF_Data(map) + offsets[row];
  else {
    const char *next = MF_Data(map) - offsets[row];
    if (GetVarInt(&next) > 0)
      item->s = MF_Data(map) + GetVarInt(&next);
    else
      item->s = "";
  }

  return IT_string;
}

static ItemTypes MappedBytesGetter (int row, Item_p item) {
  const int *offsets = (const int*) item->c.seq->data[1].p;
  MappedFile_p map = (MappedFile_p) item->c.seq->data[2].p;
  Sequence_p sizes = (Sequence_p) item->c.seq->data[3].p;
  Column column;
  
  column.seq = sizes;
  column.pos = -1;
  
  item->u.len = GetTypedItem(row, column, IT_int).i;
  if (offsets[row] >= 0)
    item->u.ptr = (const uint8_t*) MF_Data(map) + offsets[row];
  else {
    const char *next = MF_Data(map) - offsets[row];
    item->u.len = GetVarInt(&next);
    item->u.ptr = (const uint8_t*) MF_Data(map) + GetVarInt(&next);
  }

  return IT_bytes;
}

static Column MappedStringCol (MappedFile_p map, int rows, const char **nextp, int istext) {
  int r, memopos, *offsets;
  long colsize, colpos;
  const char *next, *limit;
  Sequence_p offseq;
  Column sizes, result;

  offseq = NewIntVec(rows, &offsets).seq;

  colsize = GetVarInt(nextp);
  colpos = colsize > 0 ? GetVarInt(nextp) : 0;

  if (colsize > 0) {
    sizes = MappedFixedCol(map, rows, nextp, 0);
    for (r = 0; r < rows; ++r) {
      offsets[r] = colpos;
      colpos += GetTypedItem(r, sizes, IT_int).i;
    }
  } else
    sizes = InitColumn(rows, PickFixedWidthGetter(0, rows, 0, 0), 0);

  colsize = GetVarInt(nextp);
  memopos = colsize > 0 ? GetVarInt(nextp) : 0;
  next = MF_Data(map) + memopos;
  limit = next + colsize;
  
  /* negated offsets point to the size/pos pair in the map */
  for (r = 0; next < limit; ++r) {
    r += (int) GetVarInt(&next);
    offsets[r] = MF_Data(map) - next; /* always < 0 */
    GetVarPair(&next);
  }

  /* data[0] points to a sequence owning the offset vector */
  /* data[1] points to that vector of subview offsets */
  /* data[2] points to the mapped file */
  /* data[3] points to the sizes seq if binary or is null if zero-terminated */
  result = InitColumn(rows, istext ? MappedStringGetter : MappedBytesGetter, 0);
  result.seq->cleaner = MappedStringCleaner;
  result.seq->data[0].p = IncSeqRef(offseq);
  result.seq->data[1].p = offsets;
  result.seq->data[2].p = IncSeqRef(map);
  result.seq->data[3].p = istext ? NULL : IncSeqRef(sizes.seq);

  return result;
}

View_p MapSubview (MappedFile_p map, int offset, View_p meta) {
  int c, cols, rows;
  const char *next;
  Column column;
  View_p result, subview;
  
  next = MF_Data(map) + offset;
  GetVarInt(&next);
  
  if (ViewSize(meta) == 0) {
    long desclen = GetVarInt(&next);
    const char *desc = next;
    next += desclen;
    meta = DescAsMeta(&desc, next);
  }
  
  rows = (int) GetVarInt(&next);
  cols = ViewSize(meta);
  
  if (cols == 0)
    return NoColumnView(rows);
  
  result = NewView(meta);
  if (rows > 0)
    for (c = 0; c < cols; ++c) {
      switch (ViewColType(result, c)) {
        
        case IT_string:
          column = MappedStringCol(map, rows, &next, 1); 
          break;
          
        case IT_bytes:
          column = MappedStringCol(map, rows, &next, 0);
          break;
          
        case IT_float:
        case IT_double:
          column = MappedFixedCol(map, rows, &next, 1);
          break;
          
        case IT_view:
          subview = GetTypedItem(c, ViewCol(meta, MC_subv), IT_view).v;
          column = MappedViewCol(map, rows, &next, subview); 
          break;
          
        default:
          column = MappedFixedCol(map, rows, &next, 0); 
          break;
      }
      SetViewColumns(result, c, 1, column);
    }

  return result;
}

static long BigEndianInt32 (const char *p) {
  const uint8_t *up = (const uint8_t*) p;
  return (p[0] << 24) | (up[1] << 16) | (up[2] << 8) | up[3];
}

View_p MappedToView (MappedFile_p map) {
  long datalen, rootoff;
  
  if (*(MF_Data(map) + MF_Length(map) - 8) != '\x80')
    return NULL;
    
  datalen = BigEndianInt32(MF_Data(map) + MF_Length(map) - 12) + 16;
  AdjustMappedFile(map, MF_Length(map) - datalen);
  rootoff = BigEndianInt32(MF_Data(map) + MF_Length(map) - 4);
  return MapSubview(map, rootoff, EmptyMetaView());
}

View_p OpenDataFile (const char *filename) {
  MappedFile_p map;
  
  map = OpenMappedFile(filename);
  if (map == NULL)
    return NULL;
    
  return MappedToView(map);
}
/* %include<getters.c>% */
/*
 * getters.c - Implementation of several simple getter functions.
 */

#include <stdlib.h>


static ItemTypes Getter_i0 (int row, Item_p item) {
  item->i = 0;
  return IT_int;
}

static ItemTypes Getter_i1 (int row, Item_p item) {
  const char *ptr = item->c.seq->data[0].p;
  item->i = (ptr[row>>3] >> (row&7)) & 1;
  return IT_int;
}

static ItemTypes Getter_i2 (int row, Item_p item) {
  const char *ptr = item->c.seq->data[0].p;
  item->i = (ptr[row>>2] >> 2*(row&3)) & 3;
  return IT_int;
}

static ItemTypes Getter_i4 (int row, Item_p item) {
  const char *ptr = item->c.seq->data[0].p;
  item->i = (ptr[row>>1] >> 4*(row&1)) & 15;
  return IT_int;
}

static ItemTypes Getter_i8 (int row, Item_p item) {
  const char *ptr = item->c.seq->data[0].p;
  item->i = (int8_t) ptr[row];
  return IT_int;
}

#if VALUES_MUST_BE_ALIGNED

static ItemTypes Getter_i16 (int row, Item_p item) {
  const uint8_t *ptr = (const uint8_t*) item->c.seq->data[0].p + row * 2;
#if _BIG_ENDIAN
  item->i = (((int8_t) ptr[0]) << 8) | ptr[1];
#else
  item->i = (((int8_t) ptr[1]) << 8) | ptr[0];
#endif
  return IT_int;
}

static ItemTypes Getter_i32 (int row, Item_p item) {
  const char *ptr = (const char*) item->c.seq->data[0].p + row * 4;
  int i;
  for (i = 0; i < 4; ++i)
    item->b[i] = ptr[i];
  return IT_int;
}

static ItemTypes Getter_i64 (int row, Item_p item) {
  const char *ptr = (const char*) item->c.seq->data[0].p + row * 8;
  int i;
  for (i = 0; i < 8; ++i)
    item->b[i] = ptr[i];
  return IT_wide;
}

static ItemTypes Getter_f32 (int row, Item_p item) {
  Getter_i32(row, item);
  return IT_float;
}

static ItemTypes Getter_f64 (int row, Item_p item) {
  Getter_i64(row, item);
  return IT_double;
}

#else

static ItemTypes Getter_i16 (int row, Item_p item) {
  const char *ptr = item->c.seq->data[0].p;
  item->i = ((short*) ptr)[row];
  return IT_int;
}

static ItemTypes Getter_i32 (int row, Item_p item) {
  const char *ptr = item->c.seq->data[0].p;
  item->i = ((const int*) ptr)[row];
  return IT_int;
}

static ItemTypes Getter_i64 (int row, Item_p item) {
  const char *ptr = item->c.seq->data[0].p;
  item->w = ((const int64_t*) ptr)[row];
  return IT_wide;
}

static ItemTypes Getter_f32 (int row, Item_p item) {
  const char *ptr = item->c.seq->data[0].p;
  item->f = ((const float*) ptr)[row];
  return IT_float;
}

static ItemTypes Getter_f64 (int row, Item_p item) {
  const char *ptr = item->c.seq->data[0].p;
  item->d = ((const double*) ptr)[row];
  return IT_double;
}

#endif

static ItemTypes Getter_i16r (int row, Item_p item) {
  const uint8_t *ptr = (const uint8_t*) item->c.seq->data[0].p + row * 2;
#if _BIG_ENDIAN
  item->i = (((int8_t) ptr[1]) << 8) | ptr[0];
#else
  item->i = (((int8_t) ptr[0]) << 8) | ptr[1];
#endif
  return IT_int;
}

static ItemTypes Getter_i32r (int row, Item_p item) {
  const char *ptr = (const char*) item->c.seq->data[0].p + row * 4;
  int i;
  for (i = 0; i < 4; ++i)
    item->b[i] = ptr[3-i];
  return IT_int;
}

static ItemTypes Getter_i64r (int row, Item_p item) {
  const char *ptr = (const char*) item->c.seq->data[0].p + row * 8;
  int i;
  for (i = 0; i < 8; ++i)
    item->b[i] = ptr[7-i];
  return IT_wide;
}

static ItemTypes Getter_f32r (int row, Item_p item) {
  Getter_i32r(row, item);
  return IT_float;
}

static ItemTypes Getter_f64r (int row, Item_p item) {
  Getter_i64r(row, item);
  return IT_double;
}

Getter PickIntGetter (int bits) {
  switch (bits) {
    default:  Assert(0); /* fall through */
    case 0:   return Getter_i0;
    case 1:   return Getter_i1;
    case 2:   return Getter_i2;
    case 4:   return Getter_i4;
    case 8:   return Getter_i8;
    case 16:  return Getter_i16;
    case 32:  return Getter_i32;
    case 64:  return Getter_i64;
  }
}

Getter PickFixedWidthGetter (int bytes, int rows, int real, int flip) {
  int bits;

  static char widths[8][7] = {
    {0,-1,-1,-1,-1,-1,-1},
    {0, 8,16, 1,32, 2, 4},
    {0, 4, 8, 1,16, 2,-1},
    {0, 2, 4, 8, 1,-1,16},
    {0, 2, 4,-1, 8, 1,-1},
    {0, 1, 2, 4,-1, 8,-1},
    {0, 1, 2, 4,-1,-1, 8},
    {0, 1, 2,-1, 4,-1,-1},
  };

  bits = rows < 8 && bytes < 7 ? widths[rows][bytes] : (bytes << 3) / rows;

  switch (bits) {
    case 16:  return flip ? Getter_i16r : Getter_i16;
    case 32:  return real ? flip ? Getter_f32r : Getter_f32
                          : flip ? Getter_i32r : Getter_i32;
    case 64:  return real ? flip ? Getter_f64r : Getter_f64
                          : flip ? Getter_i64r : Getter_i64;
  }

  return PickIntGetter(bits);
}
/* %include<hash.c>% */
/*
 * hash.c - Implementation of hashing functions.
 */

#include <stdlib.h>
#include <string.h>

/*
 * hash.h - Definition of hashing functions.
 */

Column (HashCol) (ItemTypes type, Column column);
int    (HashDoFind) (View_p view, int row, View_p w, Column, Column, Column);
Column (GetHashInfo) (View_p left, View_p right, int type);
View_p (GroupCol) (View_p view, Column cols, const char *name);
Column (HashValues) (View_p view);
Column (IntersectMap) (View_p view, View_p keys);
View_p (JoinView) (View_p left, View_p right, const char *name);
int    (RowHash) (View_p view, int row);
int    (StringLookup) (const char *key, Column values);
Column (UniqMap) (View_p view);
/*
 * sorted.h - Definition of shorting functions.
 */

ItemTypes (AggregateMax) (ItemTypes type, Column column, Item_p item);
ItemTypes (AggregateMin) (ItemTypes type, Column column, Item_p item);
ItemTypes (AggregateSum) (ItemTypes type, Column column, Item_p item);
int       (RowCompare) (int r1, View_p v1, int r2, View_p v2);
int       (RowEqual) (int r1, View_p v1, int r2, View_p v2);
Column    (SortMap) (View_p view);
/*
 * indirect.h - Definition of some virtual views.
 */

View_p (BlockedView) (View_p view);
View_p (ConcatView) (View_p view1, View_p view2);
View_p (GroupedView) (View_p view, Column scol, Column gcol, const char *name);
View_p (RemapSubview) (View_p view, Column mapcol, int start, int count);
View_p (StepView) (View_p view, int count, int offset, int rate, int step);
View_p (UngroupView) (View_p view, int col);

typedef struct HashInfo *HashInfo_p;

typedef struct HashInfo {
  View_p  view;     /* input view */
  int     prime;    /* prime used for hashing */
  int     fill;     /* used to fill map */
  int    *map;      /* map of unique rows */
  Column  mapcol;   /* owner of map */
  int    *hvec;     /* hash probe vector */
  Column  veccol;   /* owner of hvec */
  int    *hashes;   /* hash values, one per view row */
  Column  hashcol;  /* owner of hashes */
} HashInfo;

static int StringHash (const char *s, int n) {
  /* similar to Python's stringobject.c */
  int i, h = (*s * 0xFF) << 7;
  if (n == INVALID_COUNT)
    n = strlen(s);
  for (i = 0; i < n; ++i)
    h = (1000003 * h) ^ s[i];
  return h ^ i;
}

Column HashCol (ItemTypes type, Column column) {
  int i, count, *data;
  Column result;
  Item item;

  if (type == IT_int && column.seq->getter == PickIntGetter(32))
    return column;

  count = column.seq->count;
  result = NewIntVec(count, &data);

  switch (type) {
    
    case IT_int:
      for (i = 0; i < count; ++i)
        data[i] = GetTypedItem(i, column, IT_int).i;
      break;
        
    case IT_wide:
      for (i = 0; i < count; ++i) {
        item = GetTypedItem(i, column, IT_wide);
        data[i] = item.q[0] ^ item.q[1];
      }
      break;
        
    case IT_float:
      for (i = 0; i < count; ++i)
        data[i] = GetTypedItem(i, column, IT_float).i;
      break;
        
    case IT_double:
      for (i = 0; i < count; ++i) {
        item = GetTypedItem(i, column, IT_double);
        data[i] = item.q[0] ^ item.q[1];
      }
      break;
        
    case IT_string:
      for (i = 0; i < count; ++i) {
        item = GetTypedItem(i, column, IT_string);
        data[i] = StringHash(item.s, INVALID_COUNT);
      }
      break;
        
    case IT_bytes:
      for (i = 0; i < count; ++i) {
        item = GetTypedItem(i, column, IT_bytes);
        data[i] = StringHash((const char*) item.u.ptr, item.u.len);
      }
      break;
        
    case IT_view:
      for (i = 0; i < count; ++i) {
        int j, hcount, hval = 0;
        const int *hvec;
        Column hashes;
        
        item = GetTypedItem(i, column, IT_view);
        hashes = HashValues(item.v);
        hvec = (const int*) hashes.seq->data[0].p;
        hcount = C_Count(hashes);
        
        for (j = 0; j < hcount; ++j)
          hval ^= hvec[j];
        /* TODO: release hashes right now */
      
        data[i] = hval ^ hcount;
      }
      break;
        
    default: Assert(0);
  }

  return result;
}

int RowHash (View_p view, int row) {
  int c, cols = ViewWidth(view), hash = 0;
  Item item;

  for (c = 0; c < cols; ++c) {
    item.c = V_Cols(view)[c];
    switch (GetItem(row, &item)) {

      case IT_int:
        hash ^= item.i;
        break;

      case IT_string:
        hash ^= StringHash(item.s, INVALID_COUNT);
        break;

      case IT_bytes:
        hash ^= StringHash((const char*) item.u.ptr, item.u.len);
        break;

      default: {
        View_p rview = StepView(view, 1, row, 1, 1);
        Column rcol = HashValues(rview);
        hash = *(const int*) rcol.seq->data[0].p;
        break;
      }
    }
  }

  return hash;
}

static void XorWithIntCol (Column src, Column_p dest) {
  const int *srcdata = (const int*) src.seq->data[0].p;
  int i, count = src.seq->count, *destdata = (int*) dest->seq->data[0].p;

  for (i = 0; i < count; ++i)
    destdata[i] ^= srcdata[i];
}

Column HashValues (View_p view) {
  int c, rows = ViewSize(view), cols = ViewWidth(view);
  Column result;

  if (cols == 0 || rows == 0)
    return NewIntVec(rows, NULL);
    
  result = HashCol(ViewColType(view, 0), V_Cols(view)[0]);
  for (c = 1; c < cols; ++c) {
    Column auxcol = HashCol(ViewColType(view, c), V_Cols(view)[c]);
    /* TODO: get rid of the separate xor step by xoring in HashCol */
    XorWithIntCol(auxcol, &result);
  }
  return result;
}

static int HashFind (View_p keyview, int keyrow, int keyhash, HashInfo_p data) {
  int probe, datarow, mask, step;

  mask = data->veccol.seq->count - 1;
  probe = ~keyhash & mask;

  step = (keyhash ^ (keyhash >> 3)) & mask;
  if (step == 0)
    step = mask;

  for (;;) {
    probe = (probe + step) & mask;
    if (data->hvec[probe] == 0)
      break;

    datarow = data->map[data->hvec[probe]-1];
    if (keyhash == data->hashes[datarow] &&
        RowEqual(keyrow, keyview, datarow, data->view))
      return data->hvec[probe] - 1;

    step <<= 1;
    if (step > mask)
      step ^= data->prime;
  }

  if (keyview == data->view) {
    data->hvec[probe] = data->fill + 1;
    data->map[data->fill++] = keyrow;
  }

  return -1;
}

static int StringHashFind (const char *key, Sequence_p hseq, int prime, Column values) {
  int probe, datarow, mask, step, keyhash;
  const int *hvec = (const int*) hseq->data[0].p;

  keyhash = StringHash(key, INVALID_COUNT);
  mask = hseq->count - 1;
  probe = ~keyhash & mask;

  step = (keyhash ^ (keyhash >> 3)) & mask;
  if (step == 0)
    step = mask;

  for (;;) {
    probe = (probe + step) & mask;
    datarow = hvec[probe] - 1;
    if (datarow < 0)
      break;

    /* These string hashes are much simpler than the standard HashFind:
       no hashes vector, no indirect map, compute all hashes on-the-fly */
        
    if (strcmp(key, GetTypedItem(datarow, values, IT_string).s) == 0)
      return datarow;

    step <<= 1;
    if (step > mask)
      step ^= prime;
  }

  return ~probe;
}

static int Log2bits (int n) {
  int bits = 0;
  while ((1 << bits) < n)
    ++bits;
  return bits;
}

static Column HashVector (int rows) {
  int bits = Log2bits((4 * rows) / 3);
  if (bits < 2)
    bits = 2;
  return NewIntVec(1 << bits, NULL);
}

static void InitHashInfo (HashInfo_p info, View_p view, Column hmap, Column hvec, Column hashes) {
  int size = hvec.seq->count;

  static char slack [] = {
    0, 0, 3, 3, 3, 5, 3, 3, 29, 17, 9, 5, 83, 27, 43, 3,
    45, 9, 39, 39, 9, 5, 3, 33, 27, 9, 71, 39, 9, 5, 83, 0
  };

  info->view = view;
  info->prime = size + slack[Log2bits(size-1)];
  info->fill = 0;
  
  info->mapcol = hmap;
  info->map = (int*) hmap.seq->data[0].p;

  info->veccol = hvec;
  info->hvec = (int*) hvec.seq->data[0].p;

  info->hashcol = hashes;
  info->hashes = (int*) hashes.seq->data[0].p;
}

int StringLookup (const char *key, Column values) {
  int h, r, rows, *hptr;
  const char *string;
  Column hvec;
  HashInfo info;
  
  /* adjust data[2] and data[3], this assumes values is a string column */
  
  if (values.seq->data[2].p == NULL) {
    rows = C_Count(values);
    hvec = HashVector(rows);
    hptr = (int*) hvec.seq->data[0].p;
    
    /* use InitHashInfo to get at the prime number, bit of a hack */
    InitHashInfo(&info, NULL, hvec, hvec, hvec);
    
    for (r = 0; r < rows; ++r) {
      string = GetTypedItem(r, values, IT_string).s;
      h = StringHashFind(string, hvec.seq, info.prime, values);
      if (h < 0) /* silently ignore duplicates */
        hptr[~h] = r + 1;
    }
    
    values.seq->data[2].p = IncSeqRef(hvec.seq);
    values.seq->data[3].i = info.prime;
  }
  
  h = StringHashFind(key, (Sequence_p) values.seq->data[2].p,
                           values.seq->data[3].i, values);
  return h >= 0 ? h : -1;
}

static void FillHashInfo (HashInfo_p info, View_p view) {
  int r, rows;
  Column mapcol;
  
  rows = ViewSize(view);
  mapcol = NewIntVec(rows, NULL);   /* worst-case, don't know #groups yet */

  InitHashInfo(info, view, mapcol, HashVector(rows), HashValues(view));

  for (r = 0; r < rows; ++r)
    HashFind(view, r, info->hashes[r], info);

  /* TODO: reclaim unused entries at end of map */
  mapcol.seq->count = info->fill;
}

static void ChaseLinks (HashInfo_p info, int count, const int *hmap, const int *lmap) {
  int groups = info->fill, *smap = info->hvec, *gmap = info->hashes;
  
  while (--groups >= 0) {
    int head = hmap[groups] - 1;
    smap[groups] = count;
    while (head >= 0) {
      gmap[--count] = head;
      head = lmap[head];
    }
  }
  /* assert(count == 0); */
}

void FillGroupInfo (HashInfo_p info, View_p view) {
  int g, r, rows, *hmap, *lmap;
  Column mapcol, headmap, linkmap;
  
  rows = ViewSize(view);
  mapcol = NewIntVec(rows, NULL);   /* worst-case, don't know #groups yet */
  headmap = NewIntVec(rows, &hmap); /* worst-case, don't know #groups yet */
  linkmap = NewIntVec(rows, &lmap);

  InitHashInfo(info, view, mapcol, HashVector(rows), HashValues(view));

  for (r = 0; r < rows; ++r) {
    g = HashFind(view, r, info->hashes[r], info);
    if (g < 0)
      g = info->fill - 1;
    lmap[r] = hmap[g] - 1;
    hmap[g] = r + 1;
  }
  
  /* TODO: reclaim unused entries at end of map and hvec */
  info->mapcol.seq->count = info->veccol.seq->count = info->fill;
  
  ChaseLinks(info, rows, hmap, lmap);

  /* TODO: could release headmap and linkmap but that's a no-op here */
  
  /*  There's probably an opportunity to reduce space usage further,
      since the grouping map points to the starting row of each group:
        map[i] == gmap[smap[i]]
      Perhaps "map" (which starts out with #rows entries) can be re-used
      to append the gmap entries (if we can reduce it by #groups items).
      Or just release map[x] for grouping views, and use gmap[smap[x]].
  */
}

void GroupInfo (View_p view, Column_p mcol, Column_p scol, Column_p gcol) {
  /* TODO: get rid of this silly wrapper (due to HashInfo being private) */
  HashInfo info;
  FillGroupInfo(&info, view);
  *mcol = info.mapcol;
  *scol = info.veccol;
  *gcol = info.hashcol;
}

View_p GroupCol (View_p view, Column cols, const char *name) {
  View_p vkey, vres, gview;
  HashInfo info;
  
  vkey = ColMapView(view, cols);
  vres = ColMapView(view, OmitColumn(cols, ViewWidth(view)));

  FillGroupInfo(&info, vkey);
  gview = GroupedView(vres, info.veccol, info.hashcol, name);
  return PairView(RemapSubview(vkey, info.mapcol, 0, -1), gview);
}

static void FillJoinInfo (HashInfo_p info, View_p left, View_p right) {
  int g, r, gleft, nleft, nright, nused = 0, *hmap, *lmap, *jmap;
  Column mapcol, headmap, linkmap, joincol;
  
  nleft = ViewSize(left);
  mapcol = NewIntVec(nleft, NULL);   /* worst-case, don't know #groups yet */
  joincol = NewIntVec(nleft, &jmap);
  
  InitHashInfo(info, left, mapcol, HashVector(nleft), HashValues(left));

  for (r = 0; r < nleft; ++r) {
    g = HashFind(left, r, info->hashes[r], info);
    if (g < 0)
      g = info->fill - 1;
    jmap[r] = g;
  }

  /* TODO: reclaim unused entries at end of map */
  mapcol.seq->count = info->fill;

  gleft = info->mapcol.seq->count;
  nleft = info->hashcol.seq->count;
  nright = ViewSize(right);
  
  headmap = NewIntVec(gleft, &hmap); /* worst-case, don't know #groups yet */
  linkmap = NewIntVec(nright, &lmap);

  for (r = 0; r < nright; ++r) {
    g = HashFind(right, r, RowHash(right, r), info);
    if (g >= 0) {
      lmap[r] = hmap[g] - 1;
      hmap[g] = r + 1;
      ++nused;
    }
  }

  /* we're reusing veccol, but it might not be large enough to start with */
  /* TODO: reclaim unused entries at end of hvec */
  if (info->veccol.seq->count < nused)
    info->veccol = NewIntVec(nused, &info->hvec);
  else 
    info->veccol.seq->count = nused;

  /* reorder output to match results from FillHashInfo and FillGroupInfo */
  info->hashcol = info->veccol;
  info->hashes = info->hvec;
  info->veccol = info->mapcol;
  info->hvec = info->map;
  info->mapcol = joincol;
  info->map = jmap;
  
  ChaseLinks(info, nused, hmap, lmap);
  
  /*  As with FillGroupInfo, this is most likely not quite optimal yet.  
      All zero-length groups in smap (info->map, now info->hvec) could be 
      coalesced into one, and joinmap indices into it adjusted down a bit.
      Would reduce the size of smap when there are lots of failed matches.
      Also: FillJoinInfo needs quite a lot of temp vector space right now.
   */
}

View_p JoinView (View_p left, View_p right, const char *name) {
  View_p lnames, rnames, lkey, rkey, rres, gview;
  Column lmap, rmap;
  HashInfo info;
  
  lnames = NameColView(left);
  rnames = NameColView(right);
  lmap = IntersectMap(rnames, lnames);
  rmap = IntersectMap(lnames, rnames);
  
  lkey = ColMapView(left, lmap);
  rkey = ColMapView(right, rmap);
  rres = ColMapView(right, OmitColumn(rmap, ViewWidth(right)));
  
  FillJoinInfo(&info, lkey, rkey);
  gview = GroupedView(rres, info.veccol, info.hashcol, name);
  return PairView(left, RemapSubview(gview, info.mapcol, 0, -1));
}

Column UniqMap (View_p view) {
  HashInfo info;
  FillHashInfo(&info, view);
  return info.mapcol;
}

Column IntersectMap (View_p view, View_p keys) {
  int r, rows, *output, fill = 0;
  HashInfo info;
  Column result;
  
  FillHashInfo(&info, view);
  
  rows = ViewSize(keys);
  result = NewIntVec(rows, &output);
  
  for (r = 0; r < rows; ++r)
    if (HashFind(keys, r, RowHash(keys, r), &info) >= 0)
      output[fill++] = r;

  /* TODO: reclaim unused entries at end of map */
  result.seq->count = fill;
  return result;
}

int HashDoFind (View_p view, int row, View_p w, Column a, Column b, Column c) {
  HashInfo info;
  /* TODO: avoid Log2bits call in InitHashInfo, since it's done on each find */
  InitHashInfo(&info, w, a, b, c);
  return HashFind(view, row, RowHash(view, row), &info);
}

Column GetHashInfo (View_p left, View_p right, int type) {
  HashInfo info;
  Sequence_p seqvec[3];

  switch (type) {
    case 0:   FillHashInfo(&info, left); break;
    case 1:   FillGroupInfo(&info, left); break;
    default:  FillJoinInfo(&info, left, right); break;
  }
    
  seqvec[0] = info.mapcol.seq;
  seqvec[1] = info.veccol.seq;
  seqvec[2] = info.hashcol.seq;

  return NewSequenceColumn(IT_column, seqvec, 3);
}
/* %include<indirect.c>% */
/*
 * indirect.c - Implementation of some virtual views.
 */

#include <stdlib.h>
#include <string.h>

/*
 * mutable.h - Definition of mutable views.
 */

View_p (ViewReplace) (View_p view, int row, int count, View_p data);
View_p (ViewSet) (View_p view, int row, int col, Item_p value);

static void RemapCleaner (Sequence_p seq) {
  DecViewRef((View_p) seq->data[0].p);
  DecSeqRef((Sequence_p) seq->data[1].p);
}

static ItemTypes RemapGetter (int row, Item_p item) {
  int start;
  const int* data;
  View_p parent;

  parent = (View_p) item->c.seq->data[0].p;
  data = (const int*) ((Sequence_p) item->c.seq->data[1].p)->data[0].p;
  start = item->c.seq->data[2].i;

  item->c = V_Cols(parent)[item->c.pos];
  row += start;
  if (data[row] < 0)
    row += data[row];
  return GetItem(data[row], item);
}

View_p RemapSubview (View_p view, Column mapcol, int start, int count) {
  Sequence_p seq;

  if (count < 0)
    count = C_Count(mapcol);
  
  if (ViewWidth(view) == 0)
    return NoColumnView(count);
    
  seq = InitColumn(count, RemapGetter, 0).seq;
  seq->cleaner = RemapCleaner;
  /* data[0] is the parent view to which the map applies */
  /* data[1] is the map, as a sequence */
  /* data[2] is the index offset */
  seq->data[0].p = IncViewRef(view);
  seq->data[1].p = IncSeqRef(mapcol.seq);
  seq->data[2].i = start;

  return IndirectView(V_Meta(view), seq);
}

static View_p MakeMetaSubview (const char *name, View_p view) {
  View_p meta, result;
  Column namecol, typecol, subvcol;
  
  namecol = NewStringColumn(1, strlen(name) + 1, 1);
  AppendToStringColumn(name, -1, &namecol);

  typecol = NewStringColumn(1, 2, 1);
  AppendToStringColumn("V", -1, &typecol);
  
  meta = V_Meta(view);
  subvcol = NewSequenceColumn(IT_view, &meta, 1);
  
  result = NewView(V_Meta(meta));
  SetViewColumns(result, MC_name, 1, namecol);
  SetViewColumns(result, MC_type, 1, typecol);
  SetViewColumns(result, MC_subv, 1, subvcol);
  return result;
}

static void GroupedCleaner (Sequence_p seq) {
  DecViewRef((View_p) seq->data[0].p);
  DecSeqRef((Sequence_p) seq->data[1].p);
  DecSeqRef((Sequence_p) seq->data[2].p);
  DecSeqRef((Sequence_p) seq->data[3].p);
}

static ItemTypes GroupedGetter (int row, Item_p item) {
  View_p *subviews = ((Sequence_p) item->c.seq->data[3].p)->data[0].p;
  
  if (subviews[row] == NULL) {
    int start;
    const int *sptr;
    View_p parent;
    Sequence_p gmap;
  
    parent = item->c.seq->data[0].p;
    sptr = ((Sequence_p) item->c.seq->data[1].p)->data[0].p;
    gmap = item->c.seq->data[2].p;

    start = row > 0 ? sptr[row-1] : 0;
    item->c.seq = gmap;
    subviews[row] = IncViewRef(RemapSubview(parent, item->c,
                                            start, sptr[row] - start));
  }
  
  item->v = subviews[row];
  return IT_view;
}

View_p GroupedView (View_p view, Column startcol, Column groupcol, const char *name) {
  int groups;
  Sequence_p seq;
  Column subviewcol;
  
  groups = C_Count(startcol);
  subviewcol = NewSequenceColumn(IT_view, NULL, groups);

  seq = InitColumn(groups, GroupedGetter, 0).seq;
  seq->cleaner = GroupedCleaner;
  /* data[0] is the parent view to which the grouping applies */
  /* data[1] is the start map, as a sequence */
  /* data[2] is the group map, as a sequence */
  /* data[3] is a cache of subviews, as a pointer vector in a sequence */
  seq->data[0].p = IncViewRef(view);
  seq->data[1].p = IncSeqRef(startcol.seq);
  seq->data[2].p = IncSeqRef(groupcol.seq);
  seq->data[3].p = IncSeqRef(subviewcol.seq);

  return IndirectView(MakeMetaSubview(name, view), seq);
}

static void UnaryCleaner (Sequence_p seq) {
  DecViewRef((View_p) seq->data[0].p);
}

static ItemTypes StepGetter (int row, Item_p item) {
  int rows, offs, rate, step;
  View_p parent;
  Sequence_p seq;

  seq = item->c.seq;
  parent = (View_p) seq->data[0].p;
  offs = seq->data[1].i;
  rate = seq->data[2].i;
  step = seq->data[3].i;
  rows = ViewSize(parent);

  item->c = V_Cols(parent)[item->c.pos];
  row = (offs + (row / rate) * step) % rows;
  if (row < 0)
    row += rows;
  return GetItem(row, item);
}

View_p StepView (View_p view, int count, int offset, int rate, int step) {
  Sequence_p seq;

  /* prevent division by zero if input view is empty */
  if (ViewSize(view) == 0)
    return view;

  seq = InitColumn(count * rate, StepGetter, 0).seq;
  seq->cleaner = UnaryCleaner;
  /* data[0] is the parent view to which the changes apply */
  /* data[1] is the starting offset */
  /* data[2] is the rate to repeat an item */
  /* data[3] is the step to the next item */
  seq->data[0].p = IncViewRef(view);
  seq->data[1].i = offset;
  seq->data[2].i = rate;
  seq->data[3].i = step;

  return IndirectView(V_Meta(view), seq);
}

static void BinaryCleaner (Sequence_p seq) {
  DecViewRef((View_p) seq->data[0].p);
  DecViewRef((View_p) seq->data[1].p);
}

static ItemTypes ConcatGetter (int row, Item_p item) {
  int rows;
  View_p view;

  view = (View_p) item->c.seq->data[0].p;
  rows = ViewSize(view);
  if (row >= rows) {
    row -= rows;
    view = (View_p) item->c.seq->data[1].p;
  }

  item->c = V_Cols(view)[item->c.pos];
  return GetItem(row, item);
}

View_p ConcatView (View_p view1, View_p view2) {
  int rows1, rows2;
  Sequence_p seq;

  rows1 = ViewSize(view1);
  if (rows1 == 0)
    return view2;

  rows2 = ViewSize(view2);
  if (rows2 == 0)
    return view1;

  seq = InitColumn(rows1 + rows2, ConcatGetter, 0).seq;
  seq->cleaner = BinaryCleaner;
  /* data[0] is the left-hand view */
  /* data[1] is the right-hand view */
  seq->data[0].p = IncViewRef(view1);
  seq->data[1].p = IncViewRef(view2);

  return IndirectView(V_Meta(view1), seq);
}

static void UngroupCleaner (Sequence_p seq) {
  DecViewRef((View_p) seq->data[0].p);
  DecSeqRef((Sequence_p) seq->data[1].p);
}

static ItemTypes UngroupGetter (int row, Item_p item) {
  int col, subcol, subwidth, parentrow;
  const int *data;
  View_p view;

  view = (View_p) item->c.seq->data[0].p;
  data = (const int*) ((Sequence_p) item->c.seq->data[1].p)->data[0].p;
  subcol = item->c.seq->data[2].i;
  subwidth = item->c.seq->data[3].i;

  parentrow = data[row];
  if (parentrow < 0) {
    parentrow = data[row+parentrow];
    row = -data[row];
  } else
    row = 0;
  
  col = item->c.pos;

  if (subcol <= col && col < subcol + subwidth) {
    view = GetTypedItem(parentrow, ViewCol(view, subcol), IT_view).v;
    col -= subcol;
  } else {
    if (col >= subcol)
      col -= subwidth - 1;
    row = parentrow;
  }
  
  item->c = ViewCol(view, col);
  return GetItem(row, item);
}

View_p UngroupView (View_p view, int col) {
  int i, n, r, rows;
  Buffer buffer;
  View_p subview, meta, submeta, newmeta;
  Sequence_p seq;
  Column column, map;
  
  InitBuffer(&buffer);

  column = ViewCol(view, col);
  rows = C_Count(column);
  
  for (r = 0; r < rows; ++r) {
    subview = GetTypedItem(r, column, IT_view).v;
    n = ViewSize(subview);
    if (n > 0) {
      ADD_INT_TO_BUF(buffer, r);
      for (i = 1; i < n; ++i)
        ADD_INT_TO_BUF(buffer, -i);
    }
  }

  map = BufferAsIntCol(&buffer);
  
  /* result meta view replaces subview column with its actual meta view */
  meta = V_Meta(view);
  submeta = GetTypedItem(col, ViewCol(meta, MC_subv), IT_view).v;
  newmeta = ConcatView(FirstView(meta, col), submeta);
  newmeta = ConcatView(newmeta, LastView(meta, ViewSize(meta) - (col + 1)));
  
  seq = InitColumn(C_Count(map), UngroupGetter, 0).seq;
  seq->cleaner = UngroupCleaner;
  /* data[0] is the parent view */
  /* data[1] is ungroup map as a sequence */
  /* data[2] is the subview column */
  /* data[3] is the subview width */
  seq->data[0].p = IncViewRef(view);
  seq->data[1].p = IncSeqRef(map.seq);
  seq->data[2].i = col;
  seq->data[3].i = ViewSize(submeta);
  
  return IndirectView(newmeta, seq);
}

static void BlockedCleaner (Sequence_p seq) {
  DecViewRef((View_p) seq->data[0].p);
  DecSeqRef((Sequence_p) seq->data[1].p);
}

static ItemTypes BlockedGetter (int row, Item_p item) {
  int block;
  const int* data;
  View_p parent, subv;

  parent = (View_p) item->c.seq->data[0].p;
  data = (const int*) ((Sequence_p) item->c.seq->data[1].p)->data[0].p;

  for (block = 0; block + data[block] < row; ++block)
    ;

  if (row == block + data[block]) {
    row = block;
    block = ViewSize(parent) - 1;
  } else if (block > 0)
    row -= block + data[block-1];

  subv = GetTypedItem(block, ViewCol(parent, 0), IT_view).v;
  item->c = ViewCol(subv, item->c.pos);  
  return GetItem(row, item);
}

View_p BlockedView (View_p view) {
  int r, rows, *limits, tally = 0;
  View_p submeta;
  Sequence_p seq;
  Column blocks, offsets;

  /* view must have exactly one subview column */
  if (ViewWidth(view) != 1)
    return NULL;
  
  blocks = ViewCol(view, 0);
  rows = ViewSize(view);
  
  offsets = NewIntVec(rows, &limits);
  for (r = 0; r < rows; ++r) {
    tally += ViewSize(GetTypedItem(r, blocks, IT_view).v);
    limits[r] = tally;
  }
  
  seq = InitColumn(tally, BlockedGetter, 0).seq;
  seq->cleaner = BlockedCleaner;
  /* data[0] is the parent view */
  /* data[1] is a cumulative row count, as a sequence */
  seq->data[0].p = IncViewRef(view);
  seq->data[1].p = IncSeqRef(offsets.seq);
  
  submeta = GetTypedItem(0, ViewCol(V_Meta(view), MC_subv), IT_view).v;
  return IndirectView(submeta, seq);
}
/* %include<mutable.c>% */
/*
 * mutable.c - Implementation of mutable views.
 */

#include <stdlib.h>
#include <string.h>


static int BitTest(const uint8_t *bitmap, int row) {
  return bitmap[row/8] & (1 << row % 8);
}

static void BitSet(uint8_t *bitmap, int row) {
  bitmap[row/8] |= 1 << row % 8;
}

static void SettableCleaner (Sequence_p seq) {
  int r, rows;
  ItemTypes type;
  
  type = seq->data[3].i;
  rows = seq->count;
  
  switch (type) {

    case IT_string: {
      const char **p = (const char**) seq->data[0].p;
      for (r = 0; r < rows; ++r)
        if (p[r] != NULL)
          free((char*) p[r]);
      break;
    }

    case IT_bytes: {
      const Item *p = (const Item*) seq->data[0].p;
      for (r = 0; r < rows; ++r)
        if (p[r].u.ptr != NULL)
          free((char*) p[r].u.ptr);
      break;
    }

    case IT_view: {
      const View_p *p = (const View_p*) seq->data[0].p;
      for (r = 0; r < rows; ++r)
        if (p[r] != NULL)
          DecViewRef(p[r]);
      break;
    }
      
    default: break;
  }

  DecViewRef((View_p) seq->data[2].p);
}

static ItemTypes SettableGetter (int row, Item_p item) {
  ItemTypes type;
  Sequence_p seq = item->c.seq;
  const uint8_t *bitmap = seq->data[1].p;
  
  if (!BitTest(bitmap, row)) {
    View_p view = seq->data[2].p;
    item->c = ViewCol(view, item->c.pos);
    return GetItem(row, item);
  }
  
  type = seq->data[3].i;
  
  switch (type) {

    case IT_int:
      item->i = ((const int*) seq->data[0].p)[row];
      break;
    
    case IT_wide:
      item->w = ((const int64_t*) seq->data[0].p)[row];
      break;

    case IT_float:
      item->f = ((const float*) seq->data[0].p)[row];
      break;
                    
    case IT_double:
      item->d = ((const double*) seq->data[0].p)[row];
      break;
                    
    case IT_string:
      item->s = ((const char**) seq->data[0].p)[row];
      break;
                    
    case IT_bytes:
      item->u = ((const Item*) seq->data[0].p)[row].u;
      break;
                    
    case IT_view:
      item->v = ((const View_p*) seq->data[0].p)[row];
      break;
                    
    default: type = IT_unknown;
  }

  return type;
}
 
static int SettableSetter (int row, Column column, Item_p item) {
  int replace;
  void *data = column.seq->data[0].p;
  uint8_t *bitmap = column.seq->data[1].p;
  ItemTypes type = column.seq->data[3].i;
  
  replace = BitTest(bitmap, row);
  BitSet(bitmap, row);
  
  switch (type) {

    case IT_int:
      ((int*) data)[row] = item->i;
      break;
    
    case IT_wide:
      ((int64_t*) data)[row] = item->w;
      break;

    case IT_float:
      ((float*) data)[row] = item->f;
      break;
                    
    case IT_double:
      ((double*) data)[row] = item->d;
      break;
                    
    case IT_string: {
      char **p = (char**) data + row;
      if (replace)
        free(*p);
      *p = strdup(item->s);
      break;
    }
    
    case IT_bytes: {
      Item *p = (Item*) data + row;
      int n = item->u.len;
      if (replace)
        free((char*) p->u.ptr);
      p->u.ptr = memcpy(malloc(n), item->u.ptr, n);
      p->u.len = n;
      break;
    }
    
    case IT_view: {
      View_p *p = (View_p*) data + row;
      if (replace)
        DecViewRef(*p);
      *p = IncViewRef(item->v);
      break;
    }
    
    default: type = IT_unknown;
  }

  return replace;
}
 
static Column NewSettableVec (View_p view, int col) {
  int rows, width = 0, real = 0;
  ItemTypes type;
  Column column, result;
  
  rows = ViewSize(view);
  type = ViewColType(view, col);
  column = ViewCol(view, col);
  
  switch (type) {
    
    case IT_bytes:  width = sizeof (Item);
                    break;
                    
    case IT_string:
    case IT_view:   width = sizeof (void*);
                    break;
                    
    case IT_double: real = 1; /* fall through */
    case IT_wide:   width = 8;
                    break;
                    
    case IT_float:  real = 1; /* fall through */
    case IT_int:    width = 4;
                    break;
                    
    default: Assert(0);
  }
  
  result = InitColumn(rows, SettableGetter, rows * width + (rows + 7) / 8);
  result.seq->cleaner = SettableCleaner;
  
  /* data[0] points to the column data */
  /* data[1] points to the bitmap */
  /* data[2] is the original view */
  /* data[3] is the type of items in this column */
  result.seq->data[1].p = (char*) result.seq->data[0].p + rows * width;
  result.seq->data[2].p = IncViewRef(view);
  result.seq->data[3].i = type;
  
  result.pos = col;
  return result;
}

static View_p SettableView (View_p view) {
  int c, cols;
  View_p result;

  cols = ViewWidth(view);
  result = NewView(V_Meta(view));
  
  for (c = 0; c < cols; ++c)
    SetViewColumns(result, c, 1, NewSettableVec(view, c));

  return result;
}

static void MutableCleaner (Sequence_p seq) {
  DecViewRef((View_p) seq->data[1].p);
  DecSeqRef((Sequence_p) seq->data[2].p);
  DecSeqRef((Sequence_p) seq->data[3].p);
}

static int MutableSizer (View_p view) {
  /* same as ViewSizer, only exists to support IsMutable() check */
  return ViewCol(view, 0).seq->count;
}

static int ChooseMutView (Sequence_p mutseq, int *prow) {
  int index, limit, row = *prow;
  const int *perm;
  Sequence_p permseq;
  
  permseq = mutseq->data[2].p;
  perm = permseq->data[0].p;
  limit = permseq->data[1].i;
  
  index = perm[row];

  if (0 <= index && index < limit)
    row = -1;
  else if (index < 0) {
    row = perm[row+index] - limit;
    index = -index;
  } else {
    row = index - limit;
    index = 0;
  }

  *prow = row;
  return index;
}

static ItemTypes MutableGetter (int row, Item_p item) {
  int index;
  View_p view;
  
  index = ChooseMutView(item->c.seq, &row);

  if (row >= 0) {
    Sequence_p mapseq = item->c.seq->data[3].p;
    view = ((const View_p*) mapseq->data[0].p)[row];
  } else
    view = (View_p) item->c.seq->data[1].p;

  item->c = V_Cols(view)[item->c.pos];
  return GetItem(index, item);
}

View_p MutableView (View_p view) {
  int r, rows, *perm;
  Sequence_p seq;
  Column permcol;
  View_p result;

  rows = ViewSize(view);
  seq = InitColumn(rows, MutableGetter, 0).seq;
  seq->cleaner = MutableCleaner;
  
  permcol = NewIntVec(rows, &perm);
  permcol.seq->data[1].i = rows;
  
  /* data[0] tracks the number of insertions */
  /* data[1] is a settable version of the parent view */
  /* data[2] is the insert/delete permutation map, as a sequence */
  /* data[3] is a sequence column with all inserted views */
  seq->data[0].i = 0;
  seq->data[1].p = IncViewRef(SettableView(view));
  seq->data[2].p = IncSeqRef(permcol.seq);
  seq->data[3].p = IncSeqRef(NewSequenceColumn(IT_view, NULL, 0).seq);
  
  for (r = 0; r < rows; ++r)
    perm[r] = r;
    
  result = IndirectView(V_Meta(view), seq);
  result->sizer = MutableSizer; /* this also flags the view as being mutable */
  return result;
}

static int IsMutable(View_p view) {
  return view->sizer == MutableSizer;
}

View_p ViewSet (View_p view, int row, int col, Item_p item) {
  int index;
  Sequence_p mutseq;
  View_p setview;
  
  if (!IsMutable(view))
    view = MutableView(view);
  
  mutseq = ViewCol(view, 0).seq;
  index = ChooseMutView(mutseq, &row);

  if (row >= 0) {
    Sequence_p insseq = mutseq->data[3].p;
    View_p *modvec = insseq->data[0].p;
    View_p oldview = modvec[row];

    if (!IsMutable(oldview)) {
      modvec[row] = IncViewRef(MutableView(oldview));
      DecViewRef(oldview);
    }

    mutseq = ViewCol(modvec[row], 0).seq;
  }

  setview = mutseq->data[1].p;
  SettableSetter(index, ViewCol(setview, col), item);
  
  return view;
}

View_p ViewReplace (View_p view, int off, int count, View_p data) {
  int row, index, grow, rows, nrows, inscount, limit, *perm;
  Sequence_p mutseq, permseq;
  
  inscount = ViewSize(data);
  if (inscount > 0 && !ViewCompat(view, data))
    return NULL;

  if (!IsMutable(view))
    view = MutableView(view);
  
  row = off;
  mutseq = ViewCol(view, 0).seq;
  if (row < mutseq->count)
    index = ChooseMutView(mutseq, &row);
  else {
    index = row;
    row = -1;
  }
  
  permseq = (Sequence_p) mutseq->data[2].p;
  perm = permseq->data[0].p;
  limit = permseq->data[1].i;
  rows = permseq->count;

  grow = inscount - count;
  nrows = rows + grow;
  
#if 0
    /* turn insertion before row 0 of a subview into insertion in main view */
  if (grow > 0 && row >= 0 && index == 0) {
    index = row;
    row = -1;
  }
  
  if (row >= 0) {
    Sequence_p insseq = mutseq->data[3].p;
    View_p *modvec = insseq->data[0].p;
    View_p newview, oldview = modvec[row];

    newview = ViewReplace(oldview, index, count, data);
    
    modvec[row] = IncViewRef(newview);
    DecViewRef(oldview);

    mutseq = ViewCol(newview, 0).seq;
  }
#else
    /* with multiple replaces, bail out with an extra mutable layer for now */
    if (mutseq->data[0].i > 0)
      return ViewReplace(MutableView(view), off, count, data);
#endif

  if (grow != 0) {
    int *nperm;
    Column npermcol;
    
    npermcol = NewIntVec(nrows, &nperm);
    npermcol.seq->data[1].i = limit;
    
    memcpy(nperm, perm, off * sizeof (int));
    
    if (grow > 0)
      memcpy(nperm + off + grow, perm + off, (rows - off) * sizeof (int));
    else
      memcpy(nperm + off, perm + off - grow, (nrows - off) * sizeof (int));
  
    DecSeqRef(permseq);
    mutseq->data[2].p = IncSeqRef(npermcol.seq);
    
    perm = nperm;
  }
  
  if (inscount > 0) {
    int i, mods;
    Sequence_p modseq, *modptr, *nmodptr;
    Column nmodcol;

    mods = mutseq->data[0].i;
    modseq = mutseq->data[3].p;
    modptr = modseq->data[0].p;
    
    nmodcol = NewSequenceColumn(IT_view, NULL, ++mutseq->data[0].i);
    nmodptr = nmodcol.seq->data[0].p;

    memcpy(nmodptr, modptr, mods * sizeof (View_p));
    memset(modptr, 0, mods * sizeof (View_p)); /* prevent decref's later */
    nmodptr[mods] = IncViewRef(data);
      
    DecSeqRef(modseq);
    mutseq->data[3].p = IncSeqRef(nmodcol.seq);
  
    perm[off] = limit + mods;
    for (i = 1; i < inscount; ++i)
      perm[off+i] = -i;
  }

  mutseq->count = nrows;
  return view;
}
/* %include<sorted.c>% */
/*
 * sorted.c - Implementation of sorting functions.
 */

#include <stdlib.h>
#include <string.h>


static int ItemsEqual (ItemTypes type, Item a, Item b) {
  switch (type) {

    case IT_int:
      return a.i == b.i;

    case IT_wide:
      return a.w == b.w;

    case IT_float:
      return a.f == b.f;

    case IT_double:
      return a.d == b.d;

    case IT_string:
      return strcmp(a.s, b.s) == 0;

    case IT_bytes:
      return a.u.len == b.u.len && memcmp(a.u.ptr, b.u.ptr, a.u.len) == 0;

    case IT_view:
      return ViewCompare(a.v, b.v) == 0;
      
    default: Assert(0); return -1;
  }
}

int RowEqual (int r1, View_p v1, int r2, View_p v2) {
  int c, cols = ViewWidth(v1);
  ItemTypes type;
  Item item1, item2;

  for (c = 0; c < cols; ++c) {
    type = ViewColType(v1, c);
    item1 = GetTypedItem(r1, V_Cols(v1)[c], type);
    item2 = GetTypedItem(r2, V_Cols(v2)[c], type);

    if (!ItemsEqual(type, item1, item2))
      return 0;
  }

  return 1;
}

/* TODO: UTF-8 comparisons, also case-sensitive & insensitive */

static int ItemsCompare (ItemTypes type, Item a, Item b, int lower) {
  switch (type) {

    case IT_int:
      return (a.i > b.i) - (a.i < b.i);

    case IT_wide:
      return (a.w > b.w) - (a.w < b.w);

    case IT_float:
      return (a.f > b.f) - (a.f < b.f);

    case IT_double:
      return (a.d > b.d) - (a.d < b.d);

    case IT_string:
      return (lower ? strcasecmp : strcmp)(a.s, b.s);

    case IT_bytes: {
      int f;

      if (a.u.len == b.u.len)
        return memcmp(a.u.ptr, b.u.ptr, a.u.len);

      f = memcmp(a.u.ptr, b.u.ptr, a.u.len < b.u.len ? a.u.len : b.u.len);
      return f != 0 ? f : a.u.len - b.u.len;
    }
    
    case IT_view:
      return ViewCompare(a.v, b.v);
      
    default: Assert(0); return -1;
  }
}

int RowCompare (int r1, View_p v1, int r2, View_p v2) {
  int c, f, cols = ViewWidth(v1);
  ItemTypes type;
  Item item1, item2;

  for (c = 0; c < cols; ++c) {
    type = ViewColType(v1, c);
    item1 = GetTypedItem(r1, V_Cols(v1)[c], type);
    item2 = GetTypedItem(r2, V_Cols(v2)[c], type);

    f = ItemsCompare(type, item1, item2, 0);
    if (f != 0)
      return f;
  }

  return 0;
}

static int RowIsLess (View_p v, int a, int b) {
  int c, cols, f;
  ItemTypes type;
  Item va, vb;

  if (a != b) {
    cols = ViewWidth(v);
    for (c = 0; c < cols; ++c) {
      type = ViewColType(v, c);
      va = GetTypedItem(a, V_Cols(v)[c], type);
      vb = GetTypedItem(b, V_Cols(v)[c], type);

      f = ItemsCompare(type, va, vb, 0);
      if (f != 0)
        return f < 0;
    }
  }

  return a < b;
}

static int TestAndSwap (View_p v, int *a, int *b) {
  if (RowIsLess(v, *b, *a)) {
    int t = *a;
    *a = *b;
    *b = t;
    return 1;
  }
  return 0;
}

static void MergeSort (View_p v, int *ar, int nr, int *scr) {
  switch (nr) {
    case 2:
      TestAndSwap(v, ar, ar+1);
      break;
    case 3:
      TestAndSwap(v, ar, ar+1);
      if (TestAndSwap(v, ar+1, ar+2))
        TestAndSwap(v, ar, ar+1);
      break;
    case 4: /* TODO: optimize with if's */
      TestAndSwap(v, ar, ar+1);
      TestAndSwap(v, ar+2, ar+3);
      TestAndSwap(v, ar, ar+2);
      TestAndSwap(v, ar+1, ar+3);
      TestAndSwap(v, ar+1, ar+2);
      break;
      /* TODO: also special-case 5-item sort? */
    default: {
      int s1 = nr / 2, s2 = nr - s1;
      int *f1 = scr, *f2 = scr + s1, *t1 = f1 + s1, *t2 = f2 + s2;
      MergeSort(v, f1, s1, ar);
      MergeSort(v, f2, s2, ar+s1);
      for (;;)
        if (RowIsLess(v, *f1, *f2)) {
          *ar++ = *f1++;
          if (f1 >= t1) {
            while (f2 < t2)
              *ar++ = *f2++;
            break;
          }
        } else {
          *ar++ = *f2++;
          if (f2 >= t2) {
            while (f1 < t1)
              *ar++ = *f1++;
            break;
          }
        }
    }
  }
}

Column SortMap (View_p view) {
  int r, rows, *imap, *itmp;
  Column result;

  rows = ViewSize(view);

  if (rows <= 1 || ViewWidth(view) == 0)
    return NewIotaColumn(rows);

  result = NewIntVec(rows, &imap);
  NewIntVec(rows, &itmp);
  
  for (r = 0; r < rows; ++r)
    imap[r] = itmp[r] = r;

  MergeSort(view, imap, rows, itmp);

  return result;
}

ItemTypes AggregateMax (ItemTypes type, Column column, Item_p item) {
  int r, rows;
  Item temp;
  
  rows = C_Count(column);
  
  for (r = 0; r < rows; ++r) {
    temp = GetTypedItem(r, column, type);
    if (ItemsCompare (type, temp, *item, 0) > 0)
      *item = temp;
  }
  
  return type;
}

ItemTypes AggregateMin (ItemTypes type, Column column, Item_p item) {
  int r, rows;
  Item temp;
  
  rows = C_Count(column);
  
  for (r = 0; r < rows; ++r) {
    temp = GetTypedItem(r, column, type);
    if (ItemsCompare (type, temp, *item, 0) < 0)
      *item = temp;
  }
  
  return type;
}

ItemTypes AggregateSum (ItemTypes type, Column column, Item_p item) {
  int r, rows = C_Count(column);
  
  switch (type) {
    
    case IT_int:
      item->w = 0; 
      for (r = 0; r < rows; ++r)
        item->w += GetTypedItem(r, column, IT_int).i;
      return IT_wide;
      
    case IT_wide:
      item->w = 0;
      for (r = 0; r < rows; ++r)
        item->w += GetTypedItem(r, column, IT_wide).w;
      return IT_wide;
      
    case IT_float:
      item->d = 0; 
      for (r = 0; r < rows; ++r)
        item->d += GetTypedItem(r, column, IT_float).f;
      return IT_double;

    case IT_double:   
      item->d = 0; 
      for (r = 0; r < rows; ++r)
        item->d += GetTypedItem(r, column, IT_double).d;
      return IT_double;
      
    default:
      return IT_unknown;
  }
}
/* %include<view.c>% */
/*
 * view.c - Implementation of views.
 */

#include <stdlib.h>
#include <string.h>


int ViewSize (View_p view) {
  return view->sizer(view);
}

char GetColType (View_p meta, int col) {
  /* TODO: could be a #define */
  return *GetTypedItem(col, V_Cols(meta)[MC_type], IT_string).s;
}

static View_p ForceMetaView(View_p meta) {
  View_p newmeta;
  Column orignames, names;
  
  /* this code is needed so we can always do a hash lookup on the name column */
  
  orignames = ViewCol(meta, MC_name);
  names = ForceStringColumn(orignames);
  
  if (names.seq == orignames.seq)
    return meta;
    
    /* need to construct a new meta view with the adjusted names column */
  newmeta = NewView(V_Meta(meta));

  SetViewColumns(newmeta, MC_name, 1, names);
  SetViewColumns(newmeta, MC_type, 1, ViewCol(meta, MC_type));
  SetViewColumns(newmeta, MC_subv, 1, ViewCol(meta, MC_subv));

  return newmeta;
}

static int ViewSizer (View_p view) {
  Sequence_p seq = V_Cols(view)[0].seq;
  if (seq == NULL)
    return 0;
  return seq->count;
}

static void ViewCleaner (View_p view) {
  int c, cols = ViewWidth(view) + 1;
  for (c = 0; c < cols; ++c)
    DecSeqRef(V_Cols(view)[c].seq);
  DecViewRef(V_Meta(view));
}

static ItemTypes ViewGetter (int row, Item_p item) {
  item->c = V_Cols(item->c.seq)[row];
  return IT_column;
}

View_p NewView (View_p meta) {
  int c, cols, bytes;
  View_p view;

  /* a view must always be able to store at least one column */
  cols = ViewSize(meta);
  bytes = (cols + 1) * (sizeof(Column) + 1);

  view = InitColumn(cols, ViewGetter, bytes).seq;
  view->cleaner = ViewCleaner;
  view->sizer = ViewSizer;

  /* data[0] contains all the columns of this view (allocated inline) */
  /* data[1] points to the types string (also allocated inline) */
  /* data[2] is the meta view */
  V_Types(view) = (char*) (V_Cols(view) + cols + 1);
  V_Meta(view) = IncViewRef(ForceMetaView(meta));

  /* TODO: could optimize by storing the types with the meta view instead */
  for (c = 0; c < cols; ++c)
    V_Types(view)[c] = CharAsItemType(GetColType(meta, c));

  return view;
}

void SetViewColumns (View_p view, int first, int count, Column src) {
  int c;

  AdjustSeqRefs(src.seq, count);
  for (c = first; c < first + count; ++c) {
    DecSeqRef(V_Cols(view)[c].seq);
    V_Cols(view)[c] = src;
  }
}

View_p IndirectView (View_p meta, Sequence_p seq) {
  int c, cols;
  View_p result;
  Column column;

  cols = ViewSize(meta);
  column.seq = seq;
  column.pos = INVALID_COUNT; /* adjusted below */

  result = NewView(meta);
  SetViewColumns(result, 0, cols, column);

  for (c = 0; c < cols; ++c)
    V_Cols(result)[c].pos = c;

  return result;
}

View_p EmptyMetaView () {
  int bytes;
  View_p meta, subs [MC_limit];
  Column tempcol;
  Shared_p sh = GetShared();
  
  if (sh->empty != NULL)
    return sh->empty;

  /* meta is recursively defined, so we can't use NewView here */
  bytes = (MC_limit + 1) * (sizeof(Column) + 1);
  meta = InitColumn(MC_limit, ViewGetter, bytes).seq;
  meta->sizer = ViewSizer;
  meta->cleaner = ViewCleaner;
  V_Meta(meta) = IncViewRef(meta); /* circular */
  
  V_Types(meta) = (char*) (V_Cols(meta) + MC_limit + 1);
  V_Types(meta)[0] = V_Types(meta)[1] = IT_string;
  V_Types(meta)[2] = IT_view;

  /* initialize all but last columns of meta with static pointer contents */
  tempcol = NewStringColumn(MC_limit, MC_limit * 5, 1);
  AppendToStringColumn("name", -1, &tempcol);
  AppendToStringColumn("type", -1, &tempcol);
  AppendToStringColumn("subv", -1, &tempcol);
  SetViewColumns(meta, MC_name, 1, tempcol);

  tempcol = NewStringColumn(MC_limit, MC_limit * 2, 1);
  AppendToStringColumn("S", -1, &tempcol);
  AppendToStringColumn("S", -1, &tempcol);
  AppendToStringColumn("V", -1, &tempcol);
  SetViewColumns(meta, MC_type, 1, tempcol);

  /* same structure as meta but no rows */
  sh->empty = NewView(meta);

  /* initialize last column of meta, now that empty exists */
  subs[MC_name] = subs[MC_type] = subs[MC_subv] = sh->empty;
  SetViewColumns(meta, MC_subv, 1, NewSequenceColumn(IT_view, subs, MC_limit));

  return sh->empty;
}

View_p NoColumnView (rows) {
  View_p result = NewView(EmptyMetaView());
  SetViewColumns(result, 0, 1, InitColumn(rows, NULL, 0));
  return result;
}

View_p ColMapView (View_p view, Column mapcol) {
  int c, cols;
  const int *map;
  View_p result;

  map = (const int*) mapcol.seq->data[0].p;
  cols = mapcol.seq->count;

  if (cols == 0)
    return NoColumnView(ViewSize(view));

  result = NewView(RemapSubview(V_Meta(view), mapcol, 0, cols));

  for (c = 0; c < cols; ++c)
    SetViewColumns(result, c, 1, V_Cols(view)[map[c]]);

  return result;
}

View_p ColOmitView (View_p view, Column omitcol) {
  int c, d, *map;
  Column mapcol;
  
  mapcol = NewIntVec(ViewWidth(view), &map);

  /* TODO: make this robust, may have to add bounds check */
  for (c = 0; c < omitcol.seq->count; ++c)
    map[GetTypedItem(c, omitcol, IT_int).i] = 1;
  
  for (c = d = 0; c < mapcol.seq->count; ++c)
    if (!map[c])
      map[d++] = c;
      
  mapcol.seq->count = d; /* TODO: truncate unused area */
  return ColMapView(view, mapcol);
}

View_p NameColView (View_p view) {
  View_p meta, result;

  meta = V_Meta(view);
  result = NewView(StepView(meta, 1, 0, 1, 1));
  SetViewColumns(result, 0, 1, V_Cols(meta)[0]);

  return result;
}

View_p FirstView (View_p view, int count) {
  if (count >= ViewSize(view))
    return view;
  return StepView(view, count, 0, 1, 1);
}

View_p LastView (View_p view, int count) {
  int rows = ViewSize(view);
  if (count >= rows)
    return view;
  return StepView(view, count, rows - count, 1, 1);
}

View_p TakeView (View_p view, int count) {
  if (count >= 0)
    return StepView(view, count, 0, 1, 1);
  else
    return StepView(view, -count, -1, 1, -1);
}

View_p CloneView (View_p view) {
  return NewView(V_Meta(view));
}

View_p PairView (View_p view1, View_p view2) {
  int c, cols1, cols2;
  View_p meta, result;

  cols1 = ViewWidth(view1);
  if (cols1 == 0)
    return view2;
  
  cols2 = ViewWidth(view2);
  if (cols2 == 0)
    return view1;

  meta = ConcatView(V_Meta(view1), V_Meta(view2));
  result = NewView(meta);
  
  for (c = 0; c < cols1; ++c)
    SetViewColumns(result, c, 1, V_Cols(view1)[c]);
  for (c = 0; c < cols2; ++c)
    SetViewColumns(result, cols1 + c, 1, V_Cols(view2)[c]);
    
  return result;
}

void ViewStructure (View_p meta, Buffer_p buffer) {
  int c, cols = ViewSize(meta);
  
  for (c = 0; c < cols; ++c) {
    char type = GetColType(meta, c);
    
    if (type == 'V') {
      View_p submeta = GetTypedItem(c, V_Cols(meta)[MC_subv], IT_view).v;
      
      if (ViewSize(submeta) > 0) {
        ADD_CHAR_TO_BUF(*buffer, '(');
        ViewStructure(submeta, buffer);
        ADD_CHAR_TO_BUF(*buffer, ')');
        continue;
      }
    }
    ADD_CHAR_TO_BUF(*buffer, type);
  }
}

int MetaIsCompatible (View_p meta1, View_p meta2) {
  if (meta1 != meta2) {
    int c, cols = ViewSize(meta1);

    if (cols != ViewSize(meta2))
      return 0;
    
    for (c = 0; c < cols; ++c) {
      char type = GetColType(meta1, c);
    
      if (type != GetColType(meta2, c))
        return 0;
      
      if (type == 'V') {
        View_p v1 = GetTypedItem(c, V_Cols(meta1)[MC_subv], IT_view).v;
        View_p v2 = GetTypedItem(c, V_Cols(meta2)[MC_subv], IT_view).v;

        if (!MetaIsCompatible(v1, v2))
          return 0;
      }
    }
  }
  
  return 1;
}

int ViewCompare (View_p view1, View_p view2) {
  int f, r, rows1, rows2;
  
  if (view1 == view2)
    return 0;
  
  f = ViewCompare(V_Meta(view1), V_Meta(view2));
  if (f != 0)
    return f;
    
  rows1 = ViewSize(view1);
  rows2 = ViewSize(view2);
  
  for (r = 0; r < rows1; ++r) {
    if (r >= rows2)
      return 1;
  
    f = RowCompare(r, view1, r, view2);
    if (f != 0)
      return f < 0 ? -1 : 1;
  }
    
  return rows1 < rows2 ? -1 : 0;
}

static View_p MakeIotaMeta (const char *name) {
  View_p result;
  Column namecol, typecol;
  
  namecol = NewStringColumn(1, strlen(name) + 1, 1);
  AppendToStringColumn(name, -1, &namecol);

  typecol = NewStringColumn(1, 2, 1);
  AppendToStringColumn("I", -1, &typecol);
  
  result = NewView(V_Meta(EmptyMetaView()));
  SetViewColumns(result, MC_name, 1, namecol);
  SetViewColumns(result, MC_type, 1, typecol);
  SetViewColumns(result, MC_subv, 1, ViewCol(EmptyMetaView(), MC_subv));
  return result;
}

View_p TagView (View_p view, const char* name) {
  View_p tagview;
  
  tagview = NewView(MakeIotaMeta(name));
  SetViewColumns(tagview, 0, 1, NewIotaColumn(ViewSize(view)));
  
  return PairView(view, tagview);
}
/* %include<wrap_gen.c>% */
/* wrap_gen.c - generated code, do not edit */

#include <stdlib.h>

/* wrap_gen.h - generated code, do not edit */

ItemTypes (AppendCmd_VV) (Item_p a);
ItemTypes (AtCmd_VII) (Item_p a);
ItemTypes (BlockedCmd_V) (Item_p a);
ItemTypes (CloneCmd_V) (Item_p a);
ItemTypes (CoerceCmd_CS) (Item_p a);
ItemTypes (ColConvCmd_C) (Item_p a);
ItemTypes (ColMapCmd_Vn) (Item_p a);
ItemTypes (ColOmitCmd_Vn) (Item_p a);
ItemTypes (CompareCmd_VV) (Item_p a);
ItemTypes (CompatCmd_VV) (Item_p a);
ItemTypes (ConcatCmd_VV) (Item_p a);
ItemTypes (CountsCmd_VN) (Item_p a);
ItemTypes (CountsColCmd_C) (Item_p a);
ItemTypes (CountViewCmd_I) (Item_p a);
ItemTypes (DataCmd_VX) (Item_p a);
ItemTypes (DefCmd_OO) (Item_p a);
ItemTypes (DeleteCmd_VII) (Item_p a);
ItemTypes (EmitCmd_V) (Item_p a);
ItemTypes (ExceptCmd_VV) (Item_p a);
ItemTypes (FirstCmd_VI) (Item_p a);
ItemTypes (GetCmd_VX) (Item_p a);
ItemTypes (GetColCmd_VN) (Item_p a);
ItemTypes (GetInfoCmd_VVI) (Item_p a);
ItemTypes (GroupCmd_VnS) (Item_p a);
ItemTypes (GroupedCmd_ViiS) (Item_p a);
ItemTypes (GroupInfoCmd_V) (Item_p a);
ItemTypes (HashColCmd_SC) (Item_p a);
ItemTypes (HashFindCmd_VIViii) (Item_p a);
ItemTypes (HashInfoCmd_V) (Item_p a);
ItemTypes (HashViewCmd_V) (Item_p a);
ItemTypes (IjoinCmd_VV) (Item_p a);
ItemTypes (InsertCmd_VIV) (Item_p a);
ItemTypes (IntersectCmd_VV) (Item_p a);
ItemTypes (IotaCmd_I) (Item_p a);
ItemTypes (IsectMapCmd_VV) (Item_p a);
ItemTypes (JoinCmd_VVS) (Item_p a);
ItemTypes (JoinInfoCmd_VV) (Item_p a);
ItemTypes (LastCmd_VI) (Item_p a);
ItemTypes (LoadCmd_O) (Item_p a);
ItemTypes (LoopCmd_X) (Item_p a);
ItemTypes (MaxCmd_VNO) (Item_p a);
ItemTypes (MdefCmd_O) (Item_p a);
ItemTypes (MetaCmd_V) (Item_p a);
ItemTypes (MinCmd_VNO) (Item_p a);
ItemTypes (NameColCmd_V) (Item_p a);
ItemTypes (NamesCmd_V) (Item_p a);
ItemTypes (OmitMapCmd_iI) (Item_p a);
ItemTypes (OpenCmd_S) (Item_p a);
ItemTypes (PairCmd_VV) (Item_p a);
ItemTypes (ProductCmd_VV) (Item_p a);
ItemTypes (RefCmd_OX) (Item_p a);
ItemTypes (RefsCmd_O) (Item_p a);
ItemTypes (RemapCmd_Vi) (Item_p a);
ItemTypes (RemapSubCmd_ViII) (Item_p a);
ItemTypes (RenameCmd_VO) (Item_p a);
ItemTypes (RepeatCmd_VI) (Item_p a);
ItemTypes (ReplaceCmd_VIIV) (Item_p a);
ItemTypes (ReverseCmd_V) (Item_p a);
ItemTypes (RowEqCmd_IVIV) (Item_p a);
ItemTypes (RowHashCmd_VI) (Item_p a);
ItemTypes (SaveCmd_VS) (Item_p a);
ItemTypes (SetCmd_VIX) (Item_p a);
ItemTypes (SizeCmd_V) (Item_p a);
ItemTypes (SliceCmd_VIII) (Item_p a);
ItemTypes (SortCmd_V) (Item_p a);
ItemTypes (SortMapCmd_V) (Item_p a);
ItemTypes (SpreadCmd_VI) (Item_p a);
ItemTypes (StepCmd_VIIII) (Item_p a);
ItemTypes (StrLookupCmd_Ss) (Item_p a);
ItemTypes (StructDescCmd_V) (Item_p a);
ItemTypes (StructureCmd_V) (Item_p a);
ItemTypes (SumCmd_VN) (Item_p a);
ItemTypes (TagCmd_VS) (Item_p a);
ItemTypes (TakeCmd_VI) (Item_p a);
ItemTypes (ToCmd_OO) (Item_p a);
ItemTypes (TypeCmd_O) (Item_p a);
ItemTypes (TypesCmd_V) (Item_p a);
ItemTypes (UngroupCmd_VN) (Item_p a);
ItemTypes (UnionCmd_VV) (Item_p a);
ItemTypes (UniqMapCmd_V) (Item_p a);
ItemTypes (UniqueCmd_V) (Item_p a);
ItemTypes (ViewCmd_X) (Item_p a);
ItemTypes (ViewConvCmd_V) (Item_p a);
ItemTypes (WidthCmd_V) (Item_p a);
ItemTypes (WriteCmd_VS) (Item_p a);

/* end of generated code */
  
ItemTypes BlockedCmd_V (Item_p a) {
  a[0].v = BlockedView(a[0].v);
  return IT_view;
}

ItemTypes CloneCmd_V (Item_p a) {
  a[0].v = CloneView(a[0].v);
  return IT_view;
}

ItemTypes ColMapCmd_Vn (Item_p a) {
  a[0].v = ColMapView(a[0].v, a[1].c);
  return IT_view;
}

ItemTypes ColOmitCmd_Vn (Item_p a) {
  a[0].v = ColOmitView(a[0].v, a[1].c);
  return IT_view;
}

ItemTypes CompareCmd_VV (Item_p a) {
  a[0].i = ViewCompare(a[0].v, a[1].v);
  return IT_int;
}

ItemTypes CompatCmd_VV (Item_p a) {
  a[0].i = ViewCompat(a[0].v, a[1].v);
  return IT_int;
}

ItemTypes ConcatCmd_VV (Item_p a) {
  a[0].v = ConcatView(a[0].v, a[1].v);
  return IT_view;
}

ItemTypes CountsColCmd_C (Item_p a) {
  a[0].c = NewCountsColumn(a[0].c);
  return IT_column;
}

ItemTypes CountViewCmd_I (Item_p a) {
  a[0].v = NoColumnView(a[0].i);
  return IT_view;
}

ItemTypes FirstCmd_VI (Item_p a) {
  a[0].v = FirstView(a[0].v, a[1].i);
  return IT_view;
}

ItemTypes GetColCmd_VN (Item_p a) {
  a[0].c = ViewCol(a[0].v, a[1].i);
  return IT_column;
}

ItemTypes GetInfoCmd_VVI (Item_p a) {
  a[0].c = GetHashInfo(a[0].v, a[1].v, a[2].i);
  return IT_column;
}

ItemTypes GroupedCmd_ViiS (Item_p a) {
  a[0].v = GroupedView(a[0].v, a[1].c, a[2].c, a[3].s);
  return IT_view;
}

ItemTypes HashFindCmd_VIViii (Item_p a) {
  a[0].i = HashDoFind(a[0].v, a[1].i, a[2].v, a[3].c, a[4].c, a[5].c);
  return IT_int;
}

ItemTypes HashViewCmd_V (Item_p a) {
  a[0].c = HashValues(a[0].v);
  return IT_column;
}

ItemTypes IotaCmd_I (Item_p a) {
  a[0].c = NewIotaColumn(a[0].i);
  return IT_column;
}

ItemTypes IsectMapCmd_VV (Item_p a) {
  a[0].c = IntersectMap(a[0].v, a[1].v);
  return IT_column;
}

ItemTypes JoinCmd_VVS (Item_p a) {
  a[0].v = JoinView(a[0].v, a[1].v, a[2].s);
  return IT_view;
}

ItemTypes LastCmd_VI (Item_p a) {
  a[0].v = LastView(a[0].v, a[1].i);
  return IT_view;
}

ItemTypes MetaCmd_V (Item_p a) {
  a[0].v = V_Meta(a[0].v);
  return IT_view;
}

ItemTypes NameColCmd_V (Item_p a) {
  a[0].v = NameColView(a[0].v);
  return IT_view;
}

ItemTypes OmitMapCmd_iI (Item_p a) {
  a[0].c = OmitColumn(a[0].c, a[1].i);
  return IT_column;
}

ItemTypes PairCmd_VV (Item_p a) {
  a[0].v = PairView(a[0].v, a[1].v);
  return IT_view;
}

ItemTypes RemapSubCmd_ViII (Item_p a) {
  a[0].v = RemapSubview(a[0].v, a[1].c, a[2].i, a[3].i);
  return IT_view;
}

ItemTypes ReplaceCmd_VIIV (Item_p a) {
  a[0].v = ViewReplace(a[0].v, a[1].i, a[2].i, a[3].v);
  return IT_view;
}

ItemTypes RowEqCmd_IVIV (Item_p a) {
  a[0].i = RowEqual(a[0].i, a[1].v, a[2].i, a[3].v);
  return IT_int;
}

ItemTypes RowHashCmd_VI (Item_p a) {
  a[0].i = RowHash(a[0].v, a[1].i);
  return IT_int;
}

ItemTypes SizeCmd_V (Item_p a) {
  a[0].i = ViewSize(a[0].v);
  return IT_int;
}

ItemTypes SortMapCmd_V (Item_p a) {
  a[0].c = SortMap(a[0].v);
  return IT_column;
}

ItemTypes StepCmd_VIIII (Item_p a) {
  a[0].v = StepView(a[0].v, a[1].i, a[2].i, a[3].i, a[4].i);
  return IT_view;
}

ItemTypes StrLookupCmd_Ss (Item_p a) {
  a[0].i = StringLookup(a[0].s, a[1].c);
  return IT_int;
}

ItemTypes TagCmd_VS (Item_p a) {
  a[0].v = TagView(a[0].v, a[1].s);
  return IT_view;
}

ItemTypes TakeCmd_VI (Item_p a) {
  a[0].v = TakeView(a[0].v, a[1].i);
  return IT_view;
}

ItemTypes UngroupCmd_VN (Item_p a) {
  a[0].v = UngroupView(a[0].v, a[1].i);
  return IT_view;
}

ItemTypes UniqMapCmd_V (Item_p a) {
  a[0].c = UniqMap(a[0].v);
  return IT_column;
}

ItemTypes WidthCmd_V (Item_p a) {
  a[0].i = ViewWidth(a[0].v);
  return IT_int;
}

/* : append ( VV-V ) over size swap insert ; */
ItemTypes AppendCmd_VV (Item_p a) {
  ItemTypes t;
  /* over 0 */ a[2] = a[0];
  t = SizeCmd_V(a+2);
  /* swap 1 */ a[3] = a[1]; a[1] = a[2]; a[2] = a[3];
  t = InsertCmd_VIV(a);
  return IT_view;
}

/* : counts ( VN-C ) getcol countscol ; */
ItemTypes CountsCmd_VN (Item_p a) {
  ItemTypes t;
  t = GetColCmd_VN(a);
  t = CountsColCmd_C(a);
  return IT_column;
}

/* : delete ( VII-V ) 0 countview replace ; */
ItemTypes DeleteCmd_VII (Item_p a) {
  ItemTypes t;
  a[3].i = 0;
  t = CountViewCmd_I(a+3);
  t = ReplaceCmd_VIIV(a);
  return IT_view;
}

/* : except ( VV-V ) over isectmap over size omitmap remap ; */
ItemTypes ExceptCmd_VV (Item_p a) {
  ItemTypes t;
  /* over 0 */ a[2] = a[0];
  t = IsectMapCmd_VV(a+1);
  /* over 0 */ a[2] = a[0];
  t = SizeCmd_V(a+2);
  t = OmitMapCmd_iI(a+1);
  t = RemapCmd_Vi(a);
  return IT_view;
}

/* : groupinfo ( V-C ) dup 1 getinfo ; */
ItemTypes GroupInfoCmd_V (Item_p a) {
  ItemTypes t;
  /* dup 0 */ a[1] = a[0];
  a[2].i = 1;
  t = GetInfoCmd_VVI(a);
  return IT_column;
}

/* : hashinfo ( V-C ) dup 0 getinfo ; */
ItemTypes HashInfoCmd_V (Item_p a) {
  ItemTypes t;
  /* dup 0 */ a[1] = a[0];
  a[2].i = 0;
  t = GetInfoCmd_VVI(a);
  return IT_column;
}

/* : insert ( VIV-V ) 0 swap replace ; */
ItemTypes InsertCmd_VIV (Item_p a) {
  ItemTypes t;
  a[3].i = 0;
  /* swap 2 */ a[4] = a[2]; a[2] = a[3]; a[3] = a[4];
  t = ReplaceCmd_VIIV(a);
  return IT_view;
}

/* : intersect ( VV-V ) over isectmap remap ; */
ItemTypes IntersectCmd_VV (Item_p a) {
  ItemTypes t;
  /* over 0 */ a[2] = a[0];
  t = IsectMapCmd_VV(a+1);
  t = RemapCmd_Vi(a);
  return IT_view;
}

/* : joininfo ( VV-C ) 2 getinfo ; */
ItemTypes JoinInfoCmd_VV (Item_p a) {
  ItemTypes t;
  a[2].i = 2;
  t = GetInfoCmd_VVI(a);
  return IT_column;
}

/* : names ( V-C ) meta 0 getcol ; */
ItemTypes NamesCmd_V (Item_p a) {
  ItemTypes t;
  t = MetaCmd_V(a);
  a[1].i = 0;
  t = GetColCmd_VN(a);
  return IT_column;
}

/* : product ( VV-V ) over over size spread rrot swap size repeat pair ; */
ItemTypes ProductCmd_VV (Item_p a) {
  ItemTypes t;
  /* over 0 */ a[2] = a[0];
  /* over 1 */ a[3] = a[1];
  t = SizeCmd_V(a+3);
  t = SpreadCmd_VI(a+2);
  /* rrot 0 */ a[3] = a[2]; a[2] = a[1]; a[1] = a[0]; a[0] = a[3];
  /* swap 1 */ a[3] = a[1]; a[1] = a[2]; a[2] = a[3];
  t = SizeCmd_V(a+2);
  t = RepeatCmd_VI(a+1);
  t = PairCmd_VV(a);
  return IT_view;
}

/* : remap ( Vi-V ) 0 -1 remapsub ; */
ItemTypes RemapCmd_Vi (Item_p a) {
  ItemTypes t;
  a[2].i = 0;
  a[3].i = -1;
  t = RemapSubCmd_ViII(a);
  return IT_view;
}

/* : repeat ( VI-V ) over size imul 0 1 1 step ; */
ItemTypes RepeatCmd_VI (Item_p a) {
  ItemTypes t;
  /* over 0 */ a[2] = a[0];
  t = SizeCmd_V(a+2);
  /* imul 1 */ a[1].i *=  a[2].i;
  a[2].i = 0;
  a[3].i = 1;
  a[4].i = 1;
  t = StepCmd_VIIII(a);
  return IT_view;
}

/* : reverse ( V-V ) dup size -1 1 -1 step ; */
ItemTypes ReverseCmd_V (Item_p a) {
  ItemTypes t;
  /* dup 0 */ a[1] = a[0];
  t = SizeCmd_V(a+1);
  a[2].i = -1;
  a[3].i = 1;
  a[4].i = -1;
  t = StepCmd_VIIII(a);
  return IT_view;
}

/* : slice ( VIII-V ) rrot 1 swap step ; */
ItemTypes SliceCmd_VIII (Item_p a) {
  ItemTypes t;
  /* rrot 1 */ a[4] = a[3]; a[3] = a[2]; a[2] = a[1]; a[1] = a[4];
  a[4].i = 1;
  /* swap 3 */ a[5] = a[3]; a[3] = a[4]; a[4] = a[5];
  t = StepCmd_VIIII(a);
  return IT_view;
}

/* : sort ( V-V ) dup sortmap remap ; */
ItemTypes SortCmd_V (Item_p a) {
  ItemTypes t;
  /* dup 0 */ a[1] = a[0];
  t = SortMapCmd_V(a+1);
  t = RemapCmd_Vi(a);
  return IT_view;
}

/* : spread ( VI-V ) over size 0 rot 1 step ; */
ItemTypes SpreadCmd_VI (Item_p a) {
  ItemTypes t;
  /* over 0 */ a[2] = a[0];
  t = SizeCmd_V(a+2);
  a[3].i = 0;
  /* rot 1 */ a[4] = a[1]; a[1] = a[2]; a[2] = a[3]; a[3] = a[4];
  a[4].i = 1;
  t = StepCmd_VIIII(a);
  return IT_view;
}

/* : types ( V-C ) meta 1 getcol ; */
ItemTypes TypesCmd_V (Item_p a) {
  ItemTypes t;
  t = MetaCmd_V(a);
  a[1].i = 1;
  t = GetColCmd_VN(a);
  return IT_column;
}

/* : union ( VV-V ) over except concat ; */
ItemTypes UnionCmd_VV (Item_p a) {
  ItemTypes t;
  /* over 0 */ a[2] = a[0];
  t = ExceptCmd_VV(a+1);
  t = ConcatCmd_VV(a);
  return IT_view;
}

/* : unique ( V-V ) dup uniqmap remap ; */
ItemTypes UniqueCmd_V (Item_p a) {
  ItemTypes t;
  /* dup 0 */ a[1] = a[0];
  t = UniqMapCmd_V(a+1);
  t = RemapCmd_Vi(a);
  return IT_view;
}

/* end of generated code */

/* %include<cmds.c>% */
/*
 * cmds.c - Implementation of Tcl-specific operators.
 */

/*
 * ext_tcl.h - Definitions needed across the Tcl-specific code.
 */

#include <stdlib.h>
#include <string.h>


#define USE_TCL_STUBS 1
#include <tcl.h>

/* defined in colobj.c: */

extern Tcl_ObjType f_colObjType;

int         (CastObjToItem) (char type, Object_p obj, Item_p item);
Column      (CoerceColumn) (ItemTypes type, Column column);
Tcl_Obj    *(ColumnAsObj) (Column column);
int         (ColumnByName) (View_p meta, Object_p obj);
Tcl_Obj    *(ItemAsObj) (Item_p item, ItemTypes type);
Column      (ObjAsColumn) (Tcl_Obj *obj);

/* defined in ext_tcl.c: */

typedef struct SharedInfo {
  Tcl_Interp *interp;
  const struct CmdDispatch *cmds;
} SharedInfo;

typedef struct References *References_p;

#define TclInfo()   (GetShared()->info)
#define TclInterp() (TclInfo()->interp)
#define DecObjRef   Tcl_DecrRefCount

Object_p    (BufferAsTclList) (Buffer_p bp);
Tcl_Obj    *(IncObjRef) (Tcl_Obj *objPtr);
Tcl_Obj    *(NewTclList4) (Tcl_Obj *a, Tcl_Obj *b, Tcl_Obj *c, Tcl_Obj *d);

/* defined in viewobj.c: */

extern Tcl_ObjType f_viewObjType;

void        (AddDependencyObj) (Tcl_Obj *child, References_p refsPtr);
int         (AdjustCmdDef) (Tcl_Obj *cmd);
View_p      (ItemAsView) (Item_p item, ItemTypes type);
View_p      (ObjAsMetaView) (Tcl_Obj *obj);
View_p      (ObjAsView) (Tcl_Obj *obj);
Tcl_Obj    *(ViewAsObj) (View_p view);

ItemTypes AtCmd_VII (Item args[]) {
  View_p view;
  int row, col;

  view = args[0].v;
  row = args[1].i;
  col = args[2].i;

  if (col < 0 || col >= ViewWidth(view)) {
    args[0].e = EC_cioor;
    return IT_error;
  }
  if (row < 0 || row >= ViewSize(view)) {
    args[0].e = EC_rioor;
    return IT_error;
  }

  args[0].c = V_Cols(view)[col];
  return GetItem(row, args);
}

ItemTypes CoerceCmd_CS (Item args[]) {
  ItemTypes type = CharAsItemType(args[1].s[0]);
  args[0].c = CoerceColumn(type, args[0].c);
  return IT_column;
}

ItemTypes ColConvCmd_C (Item args[]) {
  return IT_column;
}

ItemTypes DataCmd_VX (Item args[]) {
  View_p meta, view;
  int c, cols, objc;
  const Object_p *objv;
  Column column;

  meta = args[0].v;
  cols = ViewSize(meta);
  
  objv = (const void*) args[1].u.ptr;
  objc = args[1].u.len;

  if (objc != cols) {
    Tcl_WrongNumArgs(TclInterp(), 0, objv, "data metaview ?columns ...?");
    return IT_unknown;
  }

  view = NewView(meta);
  for (c = 0; c < cols; ++c) {
    column = ObjAsColumn(objv[c]);
    column = CoerceColumn(ViewColType(view, c), column);
    SetViewColumns(view, c, 1, column);
  }

  args[0].v = view;
  return IT_view;
}

ItemTypes DefCmd_OO (Item args[]) {
  int argc, c, r, cols, rows;
  View_p meta, result;
  Object_p *argv, *coldata;
  Column column;
  
  if (MdefCmd_O(args) != IT_view)
    return IT_error;
  meta = args[0].v;
    
  if (Tcl_ListObjGetElements(TclInterp(), args[1].o, &argc, &argv) != TCL_OK)
    return IT_error;
    
  cols = ViewSize(meta);
  if (cols == 0) {
    if (argc != 0) {
      args[0].e = EC_cizwv;
      return IT_error;
    }
    
    args[0].v = NoColumnView(0);
    return IT_view;
  }
  
  if (argc % cols != 0) {
    args[0].e = EC_nmcw;
    return IT_error;
  }

  result = NewView(meta);
  rows = argc / cols;
  coldata = (Object_p*) malloc(rows * sizeof (Object_p));
  
  for (c = 0; c < cols; ++c) {
    for (r = 0; r < rows; ++r)
      coldata[r] = argv[c + cols * r];

    column = ObjAsColumn(Tcl_NewListObj(rows, coldata));
    column = CoerceColumn(ViewColType(result, c), column);
    SetViewColumns(result, c, 1, column);
  }
  
  free(coldata);
  args[0].v = result;
  return IT_view;
}

static Object_p GetViewRows (View_p view, int row, int rows, int tags) {
  int r, c, cols;
  View_p meta;
  Item item;
  Buffer buf;
  
  meta = V_Meta(view);
  cols = ViewSize(meta);
  InitBuffer(&buf);
  
  for (r = 0; r < rows; ++r)
    for (c = 0; c < cols; ++c) {
      if (tags) {
        item.c = V_Cols(meta)[0];
        ADD_PTR_TO_BUF(buf, ItemAsObj(&item, GetItem(c, &item)));
      }
      item.c = V_Cols(view)[c];
      ADD_PTR_TO_BUF(buf, ItemAsObj(&item, GetItem(r + row, &item)));
    }
    
  return BufferAsTclList(&buf);
}

static int ColumnNumber(Object_p obj, View_p meta) {
  int col;
  const char *name = Tcl_GetString(obj);
  
  switch (*name) {

    case '#': return -2;
    case '*': return -1;
    case '-': case '0': case '1': case '2': case '3': case '4':
              case '5': case '6': case '7': case '8': case '9':
      if (Tcl_GetIntFromObj(TclInterp(), obj, &col) != TCL_OK)
        return -3;

      if (col < 0)
        col += ViewSize(meta);
      break;

    default:
      col = StringLookup(name, V_Cols(meta)[MC_name]);
      if (col < 0)
        return -3;
  }

  return col;
}

ItemTypes GetCmd_VX (Item args[]) {
  View_p view;
  int i, objc, r, row, col, rows;
  const Object_p *objv;
  ItemTypes currtype;

  view = args[0].v;
  currtype = IT_view;
  
  objv = (const void*) args[1].u.ptr;
  objc = args[1].u.len;

  if (objc == 0) {
    args[0].o = GetViewRows(view, 0, ViewSize(view), 0);
    return IT_object;
  }
  
  for (i = 0; i < objc; ++i) {
    view = ItemAsView(&args[0], currtype);
    if (view == NULL)
      return IT_unknown;

    rows = ViewSize(view);

    if (Tcl_GetIntFromObj(0, objv[i], &row) != TCL_OK)
      switch (*Tcl_GetString(objv[i])) {
        case '@': args[0].v = V_Meta(view);
                  currtype = IT_view;
                  continue;
        case '#': args[0].i = rows;
                  currtype = IT_int;
                  continue;
        case '*': row = -1;
                  break;
        default:  Tcl_GetIntFromObj(TclInterp(), objv[i], &row);
                  return IT_unknown;
      }
    else if (row < 0)
      row += rows;

    if (++i >= objc) {
      if (row >= 0)
        args[0].o = GetViewRows(view, row, 1, 1); /* one tagged row */
      else {
        Buffer buf;
        InitBuffer(&buf);
        for (r = 0; r < rows; ++r)
          ADD_PTR_TO_BUF(buf, GetViewRows(view, r, 1, 0)); /* list of rows */
        args[0].o = BufferAsTclList(&buf);
      }
      return IT_object;
    }

    col = ColumnNumber(objv[i], V_Meta(view));

    if (row >= 0)
      switch (col) {
        default:  args[0].c = V_Cols(view)[col];
                  currtype = GetItem(row, args);
                  break;
        case -1:  args[0].o = GetViewRows(view, row, 1, 0);
                  currtype = IT_object;
                  break;
        case -2:  args[0].i = row;
                  currtype = IT_int;
                  break;
        case -3:  return IT_unknown;
      }
    else
      switch (col) {
        default:  args[0].c = V_Cols(view)[col];
                  currtype = IT_column;
                  break;
        case -1:  args[0].o = GetViewRows(view, 0, rows, 0);
                  currtype = IT_object;
                  break;
        case -2:  args[0].c = NewIotaColumn(rows);
                  currtype = IT_column;
                  break;
        case -3:  return IT_unknown;
      }
  }
  
  return currtype;
}

ItemTypes GroupCmd_VnS (Item args[]) {
  args[0].v = GroupCol(args[0].v, args[1].c, args[2].s);
  return IT_view;
}

ItemTypes HashColCmd_SC (Item args[]) {
  ItemTypes type = CharAsItemType(args[0].s[0]);
  args[0].c = HashCol(type, CoerceColumn(type, args[1].c));

  return IT_column;
}

ItemTypes IjoinCmd_VV (Item args[]) {
  View_p view;
  
  view = JoinView(args[0].v, args[1].v, "?");
  args[0].v = UngroupView(view, ViewWidth(view)-1);
  return IT_view;
}

ItemTypes MaxCmd_VNO (Item args[]) {
  char type;
  Column column;
  
  type = GetColType(V_Meta(args[0].v), args[1].i);
  column = ViewCol(args[0].v, args[1].i);
  
  if (!CastObjToItem(type, args[2].o, &args[0]))
    return IT_unknown;
    
  return AggregateMax(CharAsItemType(type), column, &args[0]);
}

ItemTypes MdefCmd_O (Item args[]) {
  View_p meta;

  meta = ObjAsMetaView(args[0].o);
  if (meta == NULL) {
    args[0].e = EC_imvd;
    return IT_error;
  }

  args[0].v = meta;
  return IT_view;
}

ItemTypes MinCmd_VNO (Item args[]) {
  char type;
  Column column;
  
  type = GetColType(V_Meta(args[0].v), args[1].i);
  column = ViewCol(args[0].v, args[1].i);
  
  if (!CastObjToItem(type, args[2].o, &args[0]))
    return IT_unknown;
    
  return AggregateMin(CharAsItemType(type), column, &args[0]);
}

ItemTypes RefsCmd_O (Item args[]) {
  args[0].i = args[0].o->refCount;
  return IT_int;
}

ItemTypes RenameCmd_VO (Item args[]) {
  int i, c, cols, objc, bytes;
  const char **newnames;
  View_p meta, view, newmeta, newview;
  Object_p *objv;
  Buffer buffer;
  Column names, tempcol;

  if (Tcl_ListObjGetElements(TclInterp(), args[1].o, &objc, &objv) != TCL_OK)
    return IT_unknown;
    
  if (objc % 2) {
    args[0].e = EC_rambe;
    return IT_error;
  }

  view = args[0].v;
  meta = V_Meta(view);
  names = ViewCol(meta, MC_name);
  cols = C_Count(names);

  /* TODO: could use the view "set" operator to simplify the code below */
  
  InitBuffer(&buffer);
  
  for (c = 0; c < cols; ++c)
    ADD_PTR_TO_BUF(buffer, GetTypedItem(c, names, IT_string).s);
    
  newnames = (const char**) BufferAsPtr(&buffer, 1);
  
  for (i = 0; i < objc; i += 2) {
    c = ColumnByName(meta, objv[i]);
    if (c < 0)
      return IT_unknown;
    newnames[c] = Tcl_GetString(objv[i+1]);
  }
  
  bytes = 0;
  for (c = 0; c < cols; ++c)
    bytes += strlen(newnames[c]) + 1;
    
  tempcol = NewStringColumn(cols, bytes, 1);
  for (c = 0; c < cols; ++c)
    AppendToStringColumn(newnames[c], -1, &tempcol);
  
  ReleaseBuffer(&buffer, 0);
  
  newmeta = NewView(V_Meta(meta));
  SetViewColumns(newmeta, MC_name, 1, tempcol);
  SetViewColumns(newmeta, MC_type, 1, ViewCol(meta, MC_type));
  SetViewColumns(newmeta, MC_subv, 1, ViewCol(meta, MC_subv));
  
  newview = NewView(newmeta);
  for (c = 0; c < cols; ++c)
    SetViewColumns(newview, c, 1, ViewCol(view, c));

  args[0].v = newview;
  return IT_view;
}

ItemTypes StructDescCmd_V (Item args[]) {
  Buffer buffer;
  
  InitBuffer(&buffer);
  MetaAsDesc(V_Meta(args[0].v), &buffer);
  args[0].o = Tcl_NewStringObj(BufferAsPtr(&buffer, 1), BufferFill(buffer));
  ReleaseBuffer(&buffer, 0);
  return IT_object;
}

ItemTypes StructureCmd_V (Item args[]) {
  Buffer buffer;
  
  InitBuffer(&buffer);
  ViewStructure(V_Meta(args[0].v), &buffer);
  args[0].o = Tcl_NewStringObj(BufferAsPtr(&buffer, 1), BufferFill(buffer));
  ReleaseBuffer(&buffer, 0);
  return IT_object;
}

ItemTypes SumCmd_VN (Item args[]) {
  char type;
  Column column;
  
  type = GetColType(V_Meta(args[0].v), args[1].i);
  column = ViewCol(args[0].v, args[1].i);
  
  return AggregateSum(CharAsItemType(type), column, &args[0]);
}

ItemTypes TypeCmd_O (Item args[]) {
  if (args[0].o->typePtr != NULL)
    args[0].s = args[0].o->typePtr->name;
  else
    args[0].s = "";
  return IT_string;
}

ItemTypes ViewConvCmd_V (Item args[]) {
  return IT_view;
}

/* %include<colobj.c>% */
/*
 * colobj.c - Implementation of column objects in Tcl.
 */


static const char *ErrorMessage(ErrorCodes code) {
  switch (code) {
    case EC_cioor: return "column index out of range";
    case EC_cizwv: return "cannot insert in zero-width view";
    case EC_imvd:  return "invalid meta view description";
    case EC_nmcw:  return "item count not a multiple of column width";
    case EC_rambe: return "rename arg count must be even";
    case EC_rioor: return "row index out of range";
  }
  return "?";
}

Tcl_Obj *ItemAsObj (Item_p item, ItemTypes type) {
  switch (type) {

    case IT_int:
      return Tcl_NewIntObj(item->i);

    case IT_wide:
      return Tcl_NewWideIntObj(item->w);

    case IT_float:
      return Tcl_NewDoubleObj(item->f);

    case IT_double:
      return Tcl_NewDoubleObj(item->d);

    case IT_string:
      return Tcl_NewStringObj(item->s, -1);
      
    case IT_bytes:
      return Tcl_NewByteArrayObj(item->u.ptr, item->u.len);

    case IT_object:
      return item->o;

    case IT_column:
      return ColumnAsObj(item->c);

    case IT_view:
      return ViewAsObj(item->v);

    case IT_error:
      Tcl_SetResult(TclInterp(), (char*) ErrorMessage(item->e), TCL_STATIC);
      return NULL;

    default:
      Tcl_Panic("vq-iao");
      return NULL;
  }
}

int ColumnByName (View_p meta, Object_p obj) {
  int colnum;
  const char *str = Tcl_GetString(obj);
  
  switch (*str) {

    case '#':
      return -2;

    case '*':
      return -3;

    case '-': case '0': case '1': case '2': case '3': case '4':
              case '5': case '6': case '7': case '8': case '9':
      if (Tcl_GetIntFromObj(NULL, obj, &colnum) != TCL_OK)
        return -4;
      if (colnum < 0)
        colnum += ViewSize(meta);
      return colnum;
      
    default:
      return StringLookup(str, ViewCol(meta, MC_name));
  }
}

int CastObjToItem (char type, Object_p obj, Item_p item) {
  switch (type) {

    case 'N':
      item->i = ColumnByName(V_Meta(item[-1].v), obj);
      return item->i >= 0;
    
    case 'n': {
      int *data, r, objc;
      Object_p *objv;
      View_p meta;
      Column column;
      
      /* TODO: could also convert via ObjAsColumn() */
      if (Tcl_ListObjGetElements(NULL, obj, &objc, &objv) != TCL_OK)
        return 0;
        
      column = NewIntVec(objc, &data);
      meta = V_Meta(item[-1].v);
      
      for (r = 0; r < objc; ++r) {
        data[r] = ColumnByName(meta, objv[r]);
        if (data[r] < 0)
          return 0;
      }
      
      item->c = column;
      break;
    }
    
    case 'I':
      return Tcl_GetIntFromObj(TclInterp(), obj, &item->i) == TCL_OK;

    case 'L':
      return Tcl_GetWideIntFromObj(TclInterp(), obj,
                                    (Tcl_WideInt*) &item->w) == TCL_OK;

    case 'F':
      if (Tcl_GetDoubleFromObj(TclInterp(), obj, &item->d) != TCL_OK)
        return 0;
      item->f = (float) item->d;
      break;

    case 'D':
      return Tcl_GetDoubleFromObj(TclInterp(), obj, &item->d) == TCL_OK;

    case 'S':
      item->s = Tcl_GetString(obj);
      break;

    case 'B':
      item->u.ptr = Tcl_GetByteArrayFromObj(obj, &item->u.len);
      break;

    case 'O':
      item->o = obj;
      break;

    case 'C':
      item->c = ObjAsColumn(obj);
      return item->c.seq->count != INVALID_COUNT;

    case 'V':
      item->v = ObjAsView(obj);
      return item->v != NULL;

    default:
      if (type < 'a' || type > 'z')
        return 0;
      item->c = CoerceColumn(CharAsItemType(type +'A'-'a'), ObjAsColumn(obj));
      break;
  }

  return 1; /* cast succeeded */
}

static void ListObjCleaner (Sequence_p seq) {
  DecObjRef((Tcl_Obj*) seq->data[0].p);
}

static ItemTypes ListObjGetter (int row, Item_p item) {
  Tcl_Obj *list;

  list = (Tcl_Obj*) item->c.seq->data[0].p;
  if (Tcl_ListObjIndex(NULL, list, row, &item->o) != TCL_OK)
    return IT_unknown;

  return IT_object;
}

static Tcl_Obj *ColAsListObj (Column column) {
  int r;
  Tcl_Obj *result, *obj;
  Item item;

  result = Tcl_NewListObj(0, NULL);
  if (column.seq != NULL)
    for (r = 0; r < column.seq->count; ++r) {
      item.c = column;
      obj = ItemAsObj(&item, GetItem(r, &item));
      /* TODO: faster is to collect as Tcl_Obj *vect, create list in one step */
      Tcl_ListObjAppendElement (NULL, result, obj);
    }
  return result;
}

#define OBJ_TO_COLUMN_P(o) ((Column*) &(o)->internalRep.twoPtrValue)

static void FreeColIntRep (Tcl_Obj *obj) {
  PUSH_KEEP_REFS
  
  DecSeqRef(OBJ_TO_COLUMN_P(obj)->seq);
  
  POP_KEEP_REFS
}

static void DupColIntRep (Tcl_Obj *src, Tcl_Obj *dup) {
  Tcl_Panic("vq-dcir");
}

static void UpdateColStrRep (Tcl_Obj *obj) {
  const char *string;
  int length;
  Tcl_Obj *list;

  PUSH_KEEP_REFS
  
  /* TODO: avoid using a list as intermediate representation */
  list = IncObjRef(ColAsListObj(*OBJ_TO_COLUMN_P(obj)));
  string = Tcl_GetStringFromObj(list, &length);

  obj->bytes = strcpy(ckalloc(length+1), string);
  obj->length = length;

  DecObjRef(list);
  
  POP_KEEP_REFS
}

static int SetColFromAnyRep (Tcl_Interp *interp, Tcl_Obj *obj) {
  int e = TCL_ERROR;
  Tcl_Obj *dup;
  Column column;

  PUSH_KEEP_REFS
  
  /* must duplicate the list because it'll be kept by ObjAsColumn */
  dup = Tcl_DuplicateObj(obj);
  column = ObjAsColumn(dup);

  if (column.seq->count == INVALID_COUNT) {
    Tcl_SetResult(interp, "invalid list", TCL_STATIC);
    goto FAIL;
  }

  if (obj->typePtr != NULL && obj->typePtr->freeIntRepProc != NULL)
    obj->typePtr->freeIntRepProc(obj);

  IncSeqRef(column.seq);
  *OBJ_TO_COLUMN_P(obj) = column;
  obj->typePtr = &f_colObjType;
  e = TCL_OK;
  
FAIL:
  POP_KEEP_REFS
  return e;
}

Tcl_ObjType f_colObjType = {
  "column", FreeColIntRep, DupColIntRep, UpdateColStrRep, SetColFromAnyRep
};

Column ObjAsColumn (Tcl_Obj *obj) {
  int length;
  Column result;

  if (obj->typePtr == &f_colObjType)
    return *OBJ_TO_COLUMN_P(obj);

  /* TODO: use Tcl_ListObjGetElements to avoid lindex calls in getter */
  if (Tcl_ListObjLength(TclInterp(), obj, &length) != TCL_OK)
    length = INVALID_COUNT;

  result = InitColumn(length, ListObjGetter, 0);
  result.seq->cleaner = ListObjCleaner;
  result.seq->data[0].p = IncObjRef(obj);

  return result;
}

Tcl_Obj *ColumnAsObj (Column column) {
  Tcl_Obj *result;

  result = Tcl_NewObj();
  Tcl_InvalidateStringRep(result);

  IncSeqRef(column.seq);
  *OBJ_TO_COLUMN_P(result) = column;
  result->typePtr = &f_colObjType;
  return result;
}

static int TotalStringLengths (Column column, int text) {
  int i, count, bytes, total = 0;
  Tcl_Obj *obj;
  Item item;

  count = column.seq->count;

  for (i = 0; i < count; ++i) {
    item.c = column;
    obj = ItemAsObj(&item, GetItem(i, &item));
    if (text)
      bytes = strlen(Tcl_GetString(obj)) + 1;
    else
      Tcl_GetByteArrayFromObj(obj, &bytes);
    total += bytes;
  }

  return total;
}

static int ItemAsInt (Item_p item, ItemTypes type) {
  if (type != IT_int) {
    item->o = ItemAsObj(item, type);
    if (item->o == NULL ||
        Tcl_GetIntFromObj(TclInterp(), item->o, &item->i) != TCL_OK)
      return -1; /* TODO: is -1 a good pick? */
  }
  return item->i;
}

static Column CoerceColType (ItemTypes from, ItemTypes to, Column column) {
  int i, count, *data;
  Tcl_Obj *obj;
  Item item;
  Column result;
  Sequence_p *seqvec;

  /* don't return here if from == to: makes sure int coercion ends up as i32 */
  /* TODO: coerces more than needed, e.g. iota gets expanded to i32 vec */

  count = column.seq->count;

  switch (to) {

    case IT_int:
      result = NewIntVec(count, &data);
      for (i = 0; i < count; ++i) {
        item.c = column;
        data[i] = ItemAsInt(&item, GetItem(i, &item));
      }
      break;

    case IT_wide:
      result = InitColumn(count, PickIntGetter(64), 8 * count);
      for (i = 0; i < count; ++i) {
        item.c = column;
        obj = ItemAsObj(&item, GetItem(i, &item));
        if (!CastObjToItem('L', obj, &item))
          Tcl_Panic("cct-L?");
        ((int64_t*) result.seq->data[0].p)[i] = item.w;
      }
      break;

    case IT_float:
      result = InitColumn(count, PickFixedWidthGetter (4, 1, 1, 0), 4 * count);
      for (i = 0; i < count; ++i) {
        item.c = column;
        obj = ItemAsObj(&item, GetItem(i, &item));
        if (!CastObjToItem('F', obj, &item))
          Tcl_Panic("cct-F?");
        ((float*) result.seq->data[0].p)[i] = item.f;
      }
      break;

    case IT_double:
      result = InitColumn(count, PickFixedWidthGetter (8, 1, 1, 0), 8 * count);
      for (i = 0; i < count; ++i) {
        item.c = column;
        obj = ItemAsObj(&item, GetItem(i, &item));
        if (!CastObjToItem('D', obj, &item))
          Tcl_Panic("cct-D?");
        ((double*) result.seq->data[0].p)[i] = item.d;
      }
      break;

    case IT_string:
      result = NewStringColumn(count, TotalStringLengths(column, 1), 1);
      for (i = 0; i < count; ++i) {
        item.c = column;
        obj = ItemAsObj(&item, GetItem(i, &item));
        AppendToStringColumn(Tcl_GetString(obj), -1, &result);
      }
      break;

    case IT_bytes:
      result = NewStringColumn(count, TotalStringLengths(column, 0), 0);
      for (i = 0; i < count; ++i) {
        item.c = column;
        obj = ItemAsObj(&item, GetItem(i, &item));
        item.u.ptr = Tcl_GetByteArrayFromObj(obj, &item.u.len);
        AppendToStringColumn(item.u.ptr, item.u.len, &result);
      }
      break;

    case IT_view:
      result = NewSequenceColumn(IT_view, NULL, count);
      seqvec = (Sequence_p*) result.seq->data[0].p;
      for (i = 0; i < count; ++i) {
        item.c = column;
        seqvec[i] = IncViewRef(ObjAsView(ItemAsObj(&item, GetItem(i, &item))));
      }
      break;

    default:
      Tcl_Panic("cct-%d?", to);
  }

  return result;
}

Column CoerceColumn (ItemTypes type, Column column) {
  Item item;

  /* determine the input column type by fetching one item */
  if (column.seq->count > 0) {
    item.c = column;
    return CoerceColType(GetItem(0, &item), type, column);
  }

  return column;
}
/* %include<loop.c>% */
/*
 * loop.c - Implementation of the Tcl-specific "loop" operator.
 */

/* TODO: code was copied from v3, needs a lot of cleanup */


  typedef struct LoopInfo {
    int row;
    Tcl_Obj* view;
    View_p v;
    Tcl_Obj* cursor;
  } LoopInfo;

  typedef struct ColTrace {
    LoopInfo* lip;
    int col;
    Tcl_Obj* name;
    int lastRow;
  } ColTrace;

static char* cursor_tracer(ClientData cd, Tcl_Interp* ip, const char *name1, const char *name2, int flags) {
  ColTrace* ct = (ColTrace*) cd;
  int row = ct->lip->row;
  if (row != ct->lastRow) {
    Tcl_Obj* o;
    if (ct->col < 0)
      o = Tcl_NewIntObj(row);
    else {
      Item item;
      item.c = ViewCol(ct->lip->v, ct->col);
      o = ItemAsObj(&item, GetItem(row, &item));
    }
    if (Tcl_ObjSetVar2(ip, ct->lip->cursor, ct->name, o, 0) == 0)
      return "cursor_tracer?";
    ct->lastRow = row;
  }
  return 0;
}

static int loop_vop(int oc, Tcl_Obj* const* ov) {
  LoopInfo li;
  const char *cur;
  Buffer result;
  int type = -1, e = TCL_OK;
  static const char *options[] = { "-where", "-index", "-collect", 0 };
  enum { eLOOP = -1, eWHERE, eINDEX, eCOLLECT };

  if (oc < 2 || oc > 4) {
    Tcl_WrongNumArgs(TclInterp(), 1, ov, "view ?arrayName? ?-type? body");
    return TCL_ERROR;
  }

  li.v = ObjAsView(ov[0]);
  if (li.v == NULL)
    return TCL_ERROR;

  --oc; ++ov;
  if (oc > 1 && *Tcl_GetString(*ov) != '-') {
    li.cursor = *ov++; --oc;
  } else
    li.cursor = Tcl_NewObj();
  cur = Tcl_GetString(IncObjRef(li.cursor));

  if (Tcl_GetIndexFromObj(0, *ov, options, "", 0, &type) == TCL_OK) {
    --oc; ++ov;
  }

  {
    int i, nr = ViewSize(li.v), nc = ViewWidth(li.v);
    ColTrace* ctab = (ColTrace*) ckalloc((nc + 1) * sizeof(ColTrace));
    InitBuffer(&result);
    for (i = 0; i <= nc; ++i) {
      ColTrace* ctp = ctab + i;
      ctp->lip = &li;
      ctp->lastRow = -1;
      if (i < nc) {
        ctp->col = i;
        ctp->name = Tcl_NewStringObj(GetTypedItem(i,
                            ViewCol(V_Meta(li.v), MC_name), IT_string).s, -1);
      } else {
        ctp->col = -1;
        ctp->name = Tcl_NewStringObj("#", -1);
      }
      e = Tcl_TraceVar2(TclInterp(), cur, Tcl_GetString(ctp->name),
                          TCL_TRACE_READS, cursor_tracer, (ClientData) ctp);
    }
    if (e == TCL_OK)
      switch (type) {
        case eLOOP:
          for (li.row = 0; li.row < nr; ++li.row) {
            /*A((*ov)->typePtr != &view_type);*/
            e = Tcl_EvalObj(TclInterp(), *ov);
            if (e == TCL_CONTINUE)
              e = TCL_OK;
            else if (e != TCL_OK) {
              if (e == TCL_BREAK)
                e = TCL_OK;
              else if (e == TCL_ERROR) {
                char msg[50];
                sprintf(msg, "\n  (\"loop\" body line %d)",
                                        TclInterp()->errorLine);
                Tcl_AddObjErrorInfo(TclInterp(), msg, -1);
              }
              break;
            }
          }
          break;
        case eWHERE:
        case eINDEX:
          for (li.row = 0; li.row < nr; ++li.row) {
            int f;
            e = Tcl_ExprBooleanObj(TclInterp(), *ov, &f);
            if (e != TCL_OK)
              break;
            if (f)
              ADD_INT_TO_BUF(result, li.row);
          }
          break;
        case eCOLLECT:
          for (li.row = 0; li.row < nr; ++li.row) {
            Tcl_Obj* o;
            e = Tcl_ExprObj(TclInterp(), *ov, &o);
            if (e != TCL_OK)
              break;
            ADD_PTR_TO_BUF(result, o);
          }
          break;
        default: Assert(0); return TCL_ERROR;
      }
    for (i = 0; i <= nc; ++i) {
      ColTrace* ctp = ctab + i;
      Tcl_UntraceVar2(TclInterp(), cur, Tcl_GetString(ctp->name),
                          TCL_TRACE_READS, cursor_tracer, (ClientData) ctp);
    }
    ckfree((char*) ctab);
    if (e == TCL_OK)
      switch (type) {
        case eWHERE: case eINDEX: {
          Column map = BufferAsIntCol(&result);
          if (type == eWHERE)
            Tcl_SetObjResult(TclInterp(), 
                ViewAsObj(RemapSubview(li.v, map, 0, -1)));
          else
            Tcl_SetObjResult(TclInterp(), ColumnAsObj(map));
          break;
        }
        case eCOLLECT: {
          Tcl_SetObjResult(TclInterp(), BufferAsTclList(&result));
          break;
        }
        default:
          ReleaseBuffer(&result, 0);
      }
  }
  Tcl_DecrRefCount(li.cursor);
  return e;
}

ItemTypes LoopCmd_X (Item args[]) {
  int objc;
  const Object_p *objv;

  objv = (const void*) args[0].u.ptr;
  objc = args[0].u.len;

  if (loop_vop(objc, objv) != TCL_OK)
    return IT_unknown;
    
  args[0].o = Tcl_GetObjResult(TclInterp());
  return IT_object;
}
/* %include<viewobj.c>% */
/*
 * viewobj.c - Implementation of view objects in Tcl.
 */


typedef struct Origin *Origin_p;

typedef struct Origin {
  Tcl_Obj      *definition; /* list from which this view can be built */
  References_p  parent;
} Origin;

int AdjustCmdDef (Tcl_Obj *cmd) {
  Tcl_Obj *origname, *newname;
  Tcl_CmdInfo cmdinfo;

  /* Use "::vlerq::blah ..." if it exists, else use "vlerq blah ...". */
  /* Could perhaps be simplified (optimized?) by using 8.5 ensembles. */
   
  if (Tcl_ListObjIndex(TclInterp(), cmd, 0, &origname) != TCL_OK)
    return TCL_ERROR;

  /* insert "::vlerq::" before the first list element */
  newname = Tcl_NewStringObj("::vlerq::", -1);
  Tcl_AppendObjToObj(newname, origname);
  
  if (Tcl_GetCommandInfo(TclInterp(), Tcl_GetString(newname), &cmdinfo))
    Tcl_ListObjReplace(NULL, cmd, 0, 1, 1, &newname);
  else {
    Tcl_Obj *buf[2];
    Tcl_DecrRefCount(newname);
    buf[0] = Tcl_NewStringObj("vlerq", -1);
    buf[1] = origname;
    Tcl_ListObjReplace(NULL, cmd, 0, 1, 2, buf);
  }
  return TCL_OK;
}

static Origin_p NewOrigin (Tcl_Obj *origin) {
  Origin_p result;

  result = malloc(sizeof (Origin));
  result->definition = IncObjRef(origin);
  result->parent = NULL;
  return result;
}

static View_p EvalIndirectView (Tcl_Obj *owner, Origin_p *origin) {
  Tcl_Obj *cmd;
  Tcl_SavedResult state;
  View_p result;
  Tcl_Interp *interp = TclInterp();

  Tcl_SaveResult(interp, &state);

  cmd = IncObjRef(Tcl_DuplicateObj(owner));
  result = NULL;

  /* def -> cmd namespace name conversion can probably be avoided in Tcl 8.5 */
  if (AdjustCmdDef(cmd) == TCL_OK) {
    int ac;
    Object_p *av;
    Tcl_ListObjGetElements(NULL, cmd, &ac, &av);
    /* don't use Tcl_EvalObjEx, it forces a string conversion */
    if (Tcl_EvalObjv(interp, ac, av, TCL_EVAL_GLOBAL) == TCL_OK)
      /* convert result to a view, may call EvalIndirectView recursively */
      result = ObjAsView(Tcl_GetObjResult(interp));
  }

  DecObjRef(cmd);

  if (result != NULL) {
    /* need to increment refcount before result obj is discarded in interp */
    IncViewRef(result);
    Tcl_RestoreResult(interp, &state);
    *origin = NewOrigin(Tcl_DuplicateObj(owner));
    return result;
  }

  Tcl_DiscardResult(&state);
  return NULL;
}

#define OBJ_TO_VIEW_P(o) (*(View_p*) &(o)->internalRep.twoPtrValue.ptr1)
#define OBJ_TO_ORIG_P(o) (*(Origin_p*) &(o)->internalRep.twoPtrValue.ptr2)

static void FreeViewIntRep (Tcl_Obj *objptr) {
  Origin_p origin = OBJ_TO_ORIG_P(objptr);
  
  PUSH_KEEP_REFS
  
  if (origin != NULL) {
    DecObjRef(origin->definition);
    free(origin);
  }

  DecViewRef(OBJ_TO_VIEW_P(objptr));
  
  POP_KEEP_REFS
}

static void DupViewIntRep (Tcl_Obj *src, Tcl_Obj *dup) {
  Tcl_Panic("vq-dvir");
}

static void UpdateViewStrRep (Tcl_Obj *objptr) {
  int length;
  const char *string;
  Tcl_Obj *tmpobj = NULL;
  Origin_p origin;

  PUSH_KEEP_REFS
  
  origin = OBJ_TO_ORIG_P(objptr);
  if (origin == NULL) {
    tmpobj = IncObjRef(ViewAsObj(OBJ_TO_VIEW_P(objptr)));
    origin = OBJ_TO_ORIG_P(tmpobj);
  }

  /* TODO: don't leave string rep lying around in origin->definition ? */
  string = Tcl_GetStringFromObj(origin->definition, &length);

  objptr->bytes = strcpy(ckalloc(length+1), string);
  objptr->length = length;

  if (tmpobj != NULL) {
    DecObjRef(tmpobj);
  }
  
  POP_KEEP_REFS
}

static int SetViewFromAnyRep (Tcl_Interp *interp, Tcl_Obj *objptr) {
  int e = TCL_ERROR, objc, rows;
  Tcl_Obj **objv;
  View_p view;
  Origin_p origin;

  PUSH_KEEP_REFS
  
  if (Tcl_ListObjGetElements(interp, objptr, &objc, &objv) != TCL_OK)
    goto FAIL;

  switch (objc) {

    case 0:
      view = IncViewRef(EmptyMetaView());
      origin = NULL;
      break;

    case 1:
      if (Tcl_GetIntFromObj(interp, objv[0], &rows) != TCL_OK)
        goto FAIL;
      view = IncViewRef(NoColumnView(rows));
      origin = NULL;
      break;

    default:
      view = EvalIndirectView(objptr, &origin);
      if (view == NULL)
        goto FAIL;
      /* view->refs was already incremented by EvalIndirectView */
  }

  /* we must have either a definition inside origin or a valid string rep */
  if (origin == NULL)
    Tcl_GetStringFromObj(objptr, NULL);

  if (objptr->typePtr != NULL && objptr->typePtr->freeIntRepProc != NULL)
    objptr->typePtr->freeIntRepProc(objptr);

  OBJ_TO_VIEW_P(objptr) = view;
  OBJ_TO_ORIG_P(objptr) = origin;
  objptr->typePtr = &f_viewObjType;
  e = TCL_OK;

FAIL:
  POP_KEEP_REFS
  return e;
}

Tcl_ObjType f_viewObjType = {
  "view", FreeViewIntRep, DupViewIntRep, UpdateViewStrRep, SetViewFromAnyRep
};

View_p ItemAsView (Item_p item, ItemTypes type) {
  if (type == IT_view)
    return item->v;
    
  return ObjAsView(ItemAsObj(item, type));
}

View_p ObjAsView (Tcl_Obj *obj) {
  if (Tcl_ConvertToType(TclInterp(), obj, &f_viewObjType) != TCL_OK)
    return NULL;
  return OBJ_TO_VIEW_P(obj);
}

static Tcl_Obj *MetaViewAsObj (View_p meta) {
  char typeCh;
  int width, rowNum;
  Tcl_Obj *result, *nameobj, *typeobj, *subvobj, *fieldobj, *childobj;
  Column names, types, subvs;
  Item item;

  names = V_Cols(meta)[MC_name];
  types = V_Cols(meta)[MC_type];
  subvs = V_Cols(meta)[MC_subv];

  result = Tcl_NewListObj(0, NULL);

  width = ViewSize(meta);
  for (rowNum = 0; rowNum < width; ++rowNum) {
    item.c = names;
    nameobj = ItemAsObj(&item, GetItem(rowNum, &item));
    item.c = types;
    typeobj = ItemAsObj(&item, GetItem(rowNum, &item));

    /* this ignores all but the first character of the type */
    typeCh = *Tcl_GetStringFromObj(typeobj, NULL);
    switch (typeCh) {

      case 'S':
        fieldobj = nameobj;
        break;

      case 'V':
        item.c = subvs;
        subvobj = ItemAsObj(&item, GetItem(rowNum, &item));
        childobj = MetaViewAsObj(ObjAsView(subvobj)); /* recursive call */
        fieldobj = NewTclList4(nameobj, childobj, NULL, NULL);
        break;

      default:
        fieldobj = Tcl_DuplicateObj(nameobj);
        Tcl_AppendToObj(fieldobj, ":", 1);
        Tcl_AppendObjToObj(fieldobj, typeobj);
        break;
    }

    Tcl_ListObjAppendElement(NULL, result, fieldobj);
  }

  return result;
}

Tcl_Obj *ViewAsObj (View_p view) {
  View_p meta = V_Meta(view);
  int c, rows = ViewSize(view), cols = ViewSize(meta);
  Tcl_Obj *descobj, *colobj, *result;

  if (meta == V_Meta(meta))
    if (rows == 0)
      descobj = Tcl_NewObj();
    else /* meta views are converted to a friendly nested-list format */
      descobj = NewTclList4(Tcl_NewStringObj("mdef", 4),
                            MetaViewAsObj(view), NULL, NULL);
  else
    if (cols == 0)
      descobj = Tcl_NewIntObj(rows);
    else {
      /* use ViewAsObj recursively to convert the meta view */
      descobj = NewTclList4(Tcl_NewStringObj("data", 4),
                            ViewAsObj(meta), NULL, NULL);
      /* then append all data as column entries */
      for (c = 0; c < cols; ++c) {
        colobj = ColumnAsObj(V_Cols(view)[c]);
        Tcl_ListObjAppendElement(NULL, descobj, colobj);
      }
    }

  result = Tcl_NewObj();
  OBJ_TO_VIEW_P(result) = IncViewRef(view);
  OBJ_TO_ORIG_P(result) = NewOrigin(descobj);
  result->typePtr = &f_viewObjType;
  
  Tcl_InvalidateStringRep(result);
  return result;
}

View_p ObjAsMetaView (Tcl_Obj *obj) {
  int r, rows, objc;
  Tcl_Obj *names, *types, *subvs, **objv, *entry, *nameobj, *subvobj;
  const char *name, *sep, *type;
  View_p view;

  if (Tcl_ListObjLength(TclInterp(), obj, &rows) != TCL_OK)
    return NULL;

  names = Tcl_NewListObj(0, 0);
  types = Tcl_NewListObj(0, 0);
  subvs = Tcl_NewListObj(0, 0);

  for (r = 0; r < rows; ++r) {
    Tcl_ListObjIndex(NULL, obj, r, &entry);
    if (Tcl_ListObjGetElements(TclInterp(), entry, &objc, &objv) != TCL_OK)
      goto FAIL;
    if (objc < 1 || objc > 2) {
      Tcl_SetResult(TclInterp(),
        "description must be of the form \"name:type\" or \"name {fields...}\"",
                     TCL_STATIC);
      goto FAIL;
    }

    name = Tcl_GetString(objv[0]);
    sep = strchr(name, ':');
    type = objc > 1 ? "V" : "S";

    if (sep != NULL) {
      nameobj= Tcl_NewStringObj(name, sep - name);
      if (sep[1] != 0)
        type = sep+1;
    } else
      nameobj = objv[0];

    if (objc > 1)
      subvobj = ViewAsObj(ObjAsMetaView(objv[1]));
    else
      subvobj = Tcl_NewObj();

    Tcl_ListObjAppendElement(NULL, names, nameobj);
    Tcl_ListObjAppendElement(NULL, types, Tcl_NewStringObj(type, -1));
    Tcl_ListObjAppendElement(NULL, subvs, subvobj);
  }

  view = NewView(V_Meta(EmptyMetaView()));
  SetViewColumns(view, MC_name, 1, CoerceColumn(IT_string, ObjAsColumn(names)));
  SetViewColumns(view, MC_type, 1, CoerceColumn(IT_string, ObjAsColumn(types)));
  SetViewColumns(view, MC_subv, 1, CoerceColumn(IT_view, ObjAsColumn(subvs)));
  return view;

FAIL:
  DecObjRef(names);
  DecObjRef(types);
  DecObjRef(subvs);
  return NULL;
}

void AddDependencyObj (Tcl_Obj *child, References_p refsPtr) {
  View_p view;
  /*Origin_p origin;*/

  view = ObjAsView(child);
  if (view == NULL)
    Tcl_Panic("vq-oad");

#if 0
OBJ_TO_ORIG_P(child) = (Origin_p) malloc(sizeof(Origin));
  origin = OBJ_TO_ORIG_P(child);
origin->definition = IncObjRef(Tcl_NewObj());
  origin->parent = refsPtr;
#endif
}
/* %include<ext_tcl.c>% */
/*
 * ext_tcl.c - Interface to the Tcl scripting language.
 */


#define TclInfo()   (GetShared()->info)

typedef struct CmdDispatch {
  const char *name, *args, *descr;
  ItemTypes (*proc) (Item_p);
} CmdDispatch;

/* stub interface code, removes the need to link with libtclstub*.a */
#if STATIC_BUILD
#define MyInitStubs(x) 1
#else
/*
 * stubs.h - Internal stub code, adapted from CritLib
 */

TclStubs               *tclStubsPtr        = NULL;
TclPlatStubs           *tclPlatStubsPtr    = NULL;
struct TclIntStubs     *tclIntStubsPtr     = NULL;
struct TclIntPlatStubs *tclIntPlatStubsPtr = NULL;

static int MyInitStubs (Tcl_Interp *ip) {

  typedef struct HeadOfInterp {
    char         *result;
    Tcl_FreeProc *freeProc;
    int           errorLine;
    TclStubs     *stubTable;
  } HeadOfInterp;

  HeadOfInterp *hoi = (HeadOfInterp*) ip;

  if (hoi->stubTable == NULL || hoi->stubTable->magic != TCL_STUB_MAGIC) {
    ip->result = "This extension requires stubs-support.";
    ip->freeProc = TCL_STATIC;
    return 0;
  }

  tclStubsPtr = hoi->stubTable;

  if (Tcl_PkgRequire(ip, "Tcl", "8.1", 0) == NULL) {
    tclStubsPtr = NULL;
    return 0;
  }

  if (tclStubsPtr->hooks != NULL) {
      tclPlatStubsPtr = tclStubsPtr->hooks->tclPlatStubs;
      tclIntStubsPtr = tclStubsPtr->hooks->tclIntStubs;
      tclIntPlatStubsPtr = tclStubsPtr->hooks->tclIntPlatStubs;
  }

  return 1;
}
#endif

#if NO_THREAD_CALLS
Shared_p GetShared () {
  static Shared data;
  return &data;
}
#define UngetShared Tcl_CreateExitHandler
#else
static Tcl_ThreadDataKey tls_data;
Shared_p GetShared () {
  return (Shared_p) Tcl_GetThreadData(&tls_data, sizeof(Shared));
}
#define UngetShared Tcl_CreateThreadExitHandler
#endif

Tcl_Obj *IncObjRef (Tcl_Obj *objPtr) {
  Tcl_IncrRefCount(objPtr);
  return objPtr;
}

Tcl_Obj *NewTclList4 (Tcl_Obj *a, Tcl_Obj *b, Tcl_Obj *c, Tcl_Obj *d) {
  Tcl_Obj *vec[4];
  vec[0] = a; vec[1] = b; vec[2] = c; vec[3] = d;
  return Tcl_NewListObj(c == 0 ? 2 : d == 0 ? 3 : 4, vec);
}

Object_p BufferAsTclList (Buffer_p bp) {
  int argc;
  Object_p result;
  
  argc = BufferFill(*bp) / sizeof (Object_p);
  result = Tcl_NewListObj(argc, (Object_p*) BufferAsPtr(bp, 1));
  ReleaseBuffer(bp, 0);
  return result;
}

void FailedAssert (const char *msg, const char *file, int line) {
  Tcl_Panic("Failed assertion at %s, line %d: %s\n", file, line, msg);
}

typedef struct References {
  Tcl_Obj       *refVar;    /* referenced variable */
  Tcl_HashTable  children;  /* set of dependent objects */
} References;

/*static void ReferencesCleaner(Tcl_Obj *varObj) {
  Tcl_UntraceVar(TclInterp(), Tcl_GetString(rv->var),
                  TCL_TRACE_WRITES | TCL_TRACE_UNSETS | TCL_GLOBAL_ONLY,
                  (Tcl_VarTraceProc*) RefTracer, varObj);
}*/

static void SetupRefCleaner (Tcl_Obj *owner, References_p refs) {
printf(" SetupRefCleaner %p %p\n",(void*)owner,(void*)refs);
}

static void InvalidateView (Tcl_Obj *child, References_p refs) {
printf(" InvalidateView %p %p\n",(void*)child,(void*)refs);
}

static void AddReferenceChild (References_p refs, Tcl_Obj *child) {
  int isNew;

  Tcl_CreateHashEntry(&refs->children, (const char*) child, &isNew);
  if (isNew)
    AddDependencyObj(child, refs);
}

static char *RefTracer (References_p refs, Tcl_Interp *interp, const char *name1, const char *name2, int flags) {
  Tcl_HashEntry *entry;
  Tcl_HashSearch search;

  if (flags && TCL_TRACE_WRITES) {
    Tcl_UntraceVar2(TclInterp(), name1, name2,
                     TCL_TRACE_WRITES | TCL_TRACE_UNSETS | TCL_GLOBAL_ONLY,
                     (Tcl_VarTraceProc*) RefTracer, refs);
  }

  /* reset all view objects to re-fetch the variable value */
  for (entry = Tcl_FirstHashEntry(&refs->children, &search);
        entry != NULL;
        entry = Tcl_NextHashEntry(&search))
    InvalidateView((Tcl_Obj*) entry->key.oneWordValue, refs);

  return NULL;
}

ItemTypes RefCmd_OX (Item args[]) {
  int objc;
  const Object_p *objv;
  View_p view;
  References_p refs;
  Tcl_Obj *arg, *result;
  
  arg = args[0].o;
  
  objv = (const void*) args[1].u.ptr;
  objc = args[1].u.len;

  result = Tcl_ObjGetVar2(TclInterp(), arg, 0,
                          TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
  if (result == NULL)
    return IT_unknown;

  view = ObjAsView(result);
  if (view == NULL)
    return IT_unknown;

  refs = (References_p) ckalloc(sizeof(References));
  refs->refVar = IncObjRef(arg);
  Tcl_InitHashTable(&refs->children, TCL_ONE_WORD_KEYS);
  SetupRefCleaner(result, refs);
  AddReferenceChild(refs, result);

  Tcl_TraceVar2(TclInterp(), Tcl_GetString(arg), 0,
                TCL_TRACE_WRITES | TCL_TRACE_UNSETS | TCL_GLOBAL_ONLY,
                (Tcl_VarTraceProc*) RefTracer, refs);

  args[0].o = result;
  return IT_object;
}

ItemTypes ToCmd_OO (Item args[]) {
  Tcl_Obj *result;

  result = Tcl_ObjSetVar2(TclInterp(), args[1].o, 0, args[0].o,
                          TCL_LEAVE_ERR_MSG);
  if (result == NULL)
    return IT_unknown;

  args[0].o = result;
  return IT_object;
}

ItemTypes OpenCmd_S (Item args[]) {
  View_p result;
  
  result = OpenDataFile(args[0].s);
  if (result == NULL) {
    Tcl_AppendResult(TclInterp(), "cannot open file: ", args[0].s, NULL);
    return IT_unknown;
  }
  
  args[0].v = result;
  return IT_view;
}

static void LoadedStringCleaner (MappedFile_p map) {
  DecObjRef((Object_p) map->data[3].p);
}

ItemTypes LoadCmd_O (Item args[]) {
  int len;
  const char *ptr;
  View_p view;
  MappedFile_p map;
  
  ptr = (const char*) Tcl_GetByteArrayFromObj(args[0].o, &len);

  map = InitMappedFile(ptr, len, LoadedStringCleaner);
  map->data[3].p = IncObjRef(args[0].o);
  
  view = MappedToView(map);
  if (view == NULL)
    return IT_unknown;
    
  args[0].v = view;
  return IT_view;
}

/* TODO: re-factor common code in WriteCmd_VS, SaveCmd_VS, and EmitCmd_V */

ItemTypes WriteCmd_VS (Item args[]) {
  int i, numitems;
  long bytes;
  Buffer buffer;
  Tcl_Channel chan;
  EmitItem_p items;
  
  chan = Tcl_GetChannel(TclInterp(), args[1].s, NULL);
  if (chan == NULL)
    return IT_unknown;
    
  if (Tcl_SetChannelOption(TclInterp(), chan, "-translation", "binary")
        != TCL_OK)
    return IT_unknown;
    
  InitBuffer(&buffer);
  
  bytes = EmitFull(args[0].v, &buffer);
  if (bytes < 0) {
    ReleaseBuffer(&buffer, 0);
    return IT_unknown;
  }
    
  numitems = BufferFill(buffer) / sizeof (EmitItem);
  items = (EmitItem_p) BufferAsPtr(&buffer, 1);
  
  for (i = 0; i < numitems; ++i) {
    Tcl_Write(chan, items[i].data, items[i].size);
    free((void*) items[i].data);
  }
  
  ReleaseBuffer(&buffer, 0);
  args[0].w = bytes;
  return IT_wide;
}

ItemTypes SaveCmd_VS (Item args[]) {
  int i, numitems;
  long bytes;
  Buffer buffer;
  Tcl_Channel chan;
  EmitItem_p items;
  
  chan = Tcl_OpenFileChannel(TclInterp(), args[1].s, "w", 0666);
  if (chan == NULL)
    return IT_unknown;
    
  if (Tcl_SetChannelOption(TclInterp(), chan, "-translation", "binary")
        != TCL_OK)
    return IT_unknown;
    
  InitBuffer(&buffer);
  
  bytes = EmitFull(args[0].v, &buffer);
  if (bytes < 0) {
    ReleaseBuffer(&buffer, 0);
    Tcl_Close(TclInterp(), chan);
    return IT_unknown;
  }
    
  numitems = BufferFill(buffer) / sizeof (EmitItem);
  items = (EmitItem_p) BufferAsPtr(&buffer, 1);
  
  for (i = 0; i < numitems; ++i) {
    Tcl_Write(chan, items[i].data, items[i].size);
    free((void*) items[i].data);
  }
  
  ReleaseBuffer(&buffer, 0);
  if (Tcl_Close(TclInterp(), chan) != TCL_OK)
    return IT_unknown;
    
  args[0].w = bytes;
  return IT_wide;
}

ItemTypes EmitCmd_V (Item args[]) {
  int i, numitems;
  long bytes;
  Buffer buffer;
  EmitItem_p items;
  uint8_t *data;
  Object_p result;
  
  InitBuffer(&buffer);
  
  bytes = EmitFull(args[0].v, &buffer);
  if (bytes < 0) {
    ReleaseBuffer(&buffer, 0);
    return IT_unknown;
  }

  result = Tcl_NewByteArrayObj(NULL, 0);
  data = Tcl_SetByteArrayLength(result, bytes);
  
  numitems = BufferFill(buffer) / sizeof (EmitItem);
  items = (EmitItem_p) BufferAsPtr(&buffer, 1);
  
  for (i = 0; i < numitems; ++i) {
    memcpy(data, items[i].data, items[i].size);
    data += items[i].size;
    free((void*) items[i].data);
  }
  
  ReleaseBuffer(&buffer, 0);
  args[0].o = result;
  return IT_object;
}

ItemTypes SetCmd_VIX (Item args[]) {
  int col, row, objc;
  const Object_p *objv;
  View_p view;
  Item item;
  
  view = args[0].v;
  row = args[1].i;
  
  objv = (const void*) args[2].u.ptr;
  objc = args[2].u.len;

  if (objc % 2 != 0) {
    Tcl_WrongNumArgs(TclInterp(), 0, objv, "set view row ?col value ...?");
    return IT_unknown;
  }

  while (objc > 0) {
    col = ColumnByName(V_Meta(view), objv[0]);
    if (col < 0)
      return IT_unknown;
      
    if (!CastObjToItem(GetColType(V_Meta(view), col), objv[1], &item))
      return IT_unknown;

    view = ViewSet(view, row, col, &item);

    objc -= 2; objv += 2;
  }

  args[0].v = view;
  return IT_view;
}

ItemTypes ViewCmd_X (Item args[]) {
  int e = TCL_OK, i, n, index, objc;
  const Object_p *objv;
  Tcl_Obj *result, *buf[10], **cvec;
  Tcl_Interp *interp = TclInterp();
  const CmdDispatch *cmds = TclInfo()->cmds;

  objv = (const void*) args[0].u.ptr;
  objc = args[0].u.len;

  if (objc < 1) {
    Tcl_WrongNumArgs(interp, 0, objv, "view arg ?op ...? ?| ...?");
    return IT_unknown;
  }
  
  Tcl_SetObjResult(interp, objv[0]); --objc; ++objv;
  
  while (e == TCL_OK && objc > 0) {
    for (n = 0; n < objc; ++n)
      if (objv[n]->bytes != 0 && *objv[n]->bytes == '|' && objv[n]->length == 1)
        break;

    if (n > 0) {
      cvec = n > 8 ? malloc((n+2) * sizeof(Tcl_Obj*)) : buf;
        
      if (Tcl_GetIndexFromObjStruct(NULL, *objv, cmds, sizeof *cmds, 
                                              "", TCL_EXACT, &index) != TCL_OK)
        index = -1;

      cvec[0] = Tcl_NewStringObj("vlerq", -1);
      cvec[1] = objv[0];
      cvec[2] = IncObjRef(Tcl_GetObjResult(interp));
      for (i = 1; i < n; ++i)
        cvec[i+2] = objv[i];
      
      result = Tcl_NewListObj(n+1, cvec+1);

      if (index < 0 || *cmds[index].args != 'V') {
        e = AdjustCmdDef(result);
        if (e == TCL_OK) {
          int ac;
          Object_p *av;
          Tcl_ListObjGetElements(NULL, result, &ac, &av);
          /* don't use Tcl_EvalObjEx, it forces a string conversion */
          e = Tcl_EvalObjv(interp, ac, av, 0);
        }
        DecObjRef(result);
      } else
        Tcl_SetObjResult(interp, result);
      
      Tcl_DecrRefCount(cvec[2]);
      if (n > 8)
        free(cvec);
    }
    
    objc -= n+1; objv += n+1; /*++k;*/
  }

  if (e != TCL_OK) {
#if 0
    char msg[50];
    sprintf(msg, "\n    (\"view\" step %d)", k);
    Tcl_AddObjErrorInfo(interp, msg, -1);
#endif
    return IT_unknown;
  }
  
  args[0].o = Tcl_GetObjResult(interp);
  return IT_object;
}

#define MAX_STACK 20

static int VlerqObjCmd (ClientData data, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[]) {
  int i, index, ok = TCL_ERROR;
  ItemTypes type;
  Tcl_Obj *result;
  Item stack [MAX_STACK];
  const char *args;
  const CmdDispatch *cmdtable;
  
  PUSH_KEEP_REFS
  
  TclInfo()->interp = interp;
  
  if (objc <= 1) {
    Tcl_WrongNumArgs(interp, 1, objv, "command ...");
    goto FAIL;
  }

  cmdtable = (const CmdDispatch*) data;
  if (Tcl_GetIndexFromObjStruct(interp, objv[1], cmdtable, sizeof *cmdtable, 
                                "command", TCL_EXACT, &index) != TCL_OK)
    goto FAIL;

  objv += 2; objc -= 2;
  args = cmdtable[index].args + 2; /* skip return type and ':' */

  for (i = 0; args[i] != 0; ++i) {
    Assert(i < MAX_STACK);
    if (args[i] == 'X') {
      Assert(args[i+1] == 0);
      stack[i].u.ptr = (const void*) (objv+i);
      stack[i].u.len = objc-i;
      break;
    }
    if ((args[i] == 0 && i != objc) || (args[i] != 0 && i >= objc)) {
      Tcl_WrongNumArgs(interp, 2, objv-2, cmdtable[index].descr);
      goto FAIL;
    }
    if (!CastObjToItem(args[i], objv[i], stack+i))
      goto FAIL; /* TODO: append info about which arg is bad */
  }
  
  TclInfo()->cmds = cmdtable; /* for ViewCmd_X */
  
  type = cmdtable[index].proc(stack);
  if (type == IT_unknown)
    goto FAIL;

  result = ItemAsObj(stack, type);
  if (result == NULL)
    goto FAIL;

  Tcl_SetObjResult(interp, result);
  ok = TCL_OK;

FAIL:
  POP_KEEP_REFS
  return ok;
}

static void DoneShared (ClientData cd) {
  Shared_p sh = (Shared_p) cd;
  if (--sh->refs == 0)
    free(sh->info);
}

static void InitShared () {
  Shared_p sh = GetShared();
  if (sh->refs++ == 0)
    sh->info = calloc(1, sizeof(SharedInfo));
  UngetShared(DoneShared, (ClientData) sh);
}

static CmdDispatch f_commands[] = {
/* opdefs_gen.h - generated code, do not edit */

/* 85 operators, using these definitions:

  # name       in     out    inline-code
  : drop       U      {}     {}
  : dup        U      UU     {$1 = $0;}
  : imul       II     I      {$0.i *=  $1.i;}
  : nip        UU     U      {$0 = $1;}
  : over       UU     UUU    {$2 = $0;}
  : rot        UUU    UUU    {$3 = $0; $0 = $1; $1 = $2; $2 = $3;}
  : rrot       UUU    UUU    {$3 = $2; $2 = $1; $1 = $0; $0 = $3;}
  : swap       UU     UU     {$2 = $0; $0 = $1; $1 = $2;}
  : tuck       UU     UUU    {$1 = $2; $0 = $1; $2 = $0;}
    
  # name       in    out
  : At         VII    U
  : Coerce     CS     U
  : ColConv    C      C
  : Data       VX     V
  : Def        OO     V
  : Emit       V      O
  : Get        VX     U
  : Group      VnS    V
  : HashCol    SC     C
  : Ijoin      VV     V
  : Load       O      V
  : Loop       X      O
  : Max        VNO    O
  : Mdef       O      V
  : Min        VNO    O
  : Open       S      V
  : Ref        OX     O
  : Refs       O      I
  : Rename     VO     V
  : Save       VS     I
  : Set        VIX    V
  : StructDesc V      S
  : Structure  V      S
  : Sum        VN     O
  : To         OO     O
  : Type       O      O
  : View       X      O
  : ViewConv   V      V
  : Write      VS     I
               
  # name       in    out  internal-name
  : Blocked    V      V   BlockedView
  : Clone      V      V   CloneView
  : ColMap     Vn     V   ColMapView
  : ColOmit    Vn     V   ColOmitView
  : Compare    VV     I   ViewCompare
  : Compat     VV     I   ViewCompat
  : Concat     VV     V   ConcatView
  : CountsCol  C      C   NewCountsColumn
  : CountView  I      V   NoColumnView
  : First      VI     V   FirstView
  : GetCol     VN     C   ViewCol
  : HashFind   VIViii I   HashDoFind
  : GetInfo    VVI    C   GetHashInfo
  : Grouped    ViiS   V   GroupedView
  : HashView   V      C   HashValues
  : IsectMap   VV     C   IntersectMap
  : Iota       I      C   NewIotaColumn
  : Join       VVS    V   JoinView
  : Last       VI     V   LastView
  : Meta       V      V   V_Meta
  : NameCol    V      V   NameColView
  : OmitMap    iI     C   OmitColumn
  : Pair       VV     V   PairView
  : RemapSub   ViII   V   RemapSubview
  : Replace    VIIV   V   ViewReplace
  : RowEq      IVIV   I   RowEqual
  : RowHash    VI     I   RowHash
  : Size       V      I   ViewSize
  : SortMap    V      C   SortMap
  : Step       VIIII  V   StepView
  : StrLookup  Ss     I   StringLookup
  : Tag        VS     V   TagView
  : Take       VI     V   TakeView
  : Ungroup    VN     V   UngroupView
  : UniqMap    V      C   UniqMap
  : Width      V      I   ViewWidth
               
  # name       in    out  compound-definition
  : Append     VV     V   {over size swap insert}
  : Counts     VN     C   {getcol countscol}
  : Delete     VII    V   {0 countview replace}
  : Except     VV     V   {over isectmap over size omitmap remap}
  : GroupInfo  V      C   {dup 1 getinfo}
  : HashInfo   V      C   {dup 0 getinfo}
  : Insert     VIV    V   {0 swap replace}
  : JoinInfo   VV     C   {2 getinfo}
  : Intersect  VV     V   {over isectmap remap}
  : Names      V      C   {meta 0 getcol}
  : Product    VV     V   {over over size spread rrot swap size repeat pair}
  : Repeat     VI     V   {over size imul 0 1 1 step}
  : Remap      Vi     V   {0 -1 remapsub}
  : Reverse    V      V   {dup size -1 1 -1 step}
  : Slice      VIII   V   {rrot 1 swap step}
  : Sort       V      V   {dup sortmap remap}
  : Spread     VI     V   {over size 0 rot 1 step}
  : Types      V      C   {meta 1 getcol}
  : Unique     V      V   {dup uniqmap remap}
  : Union      VV     V   {over except concat}
  
  # some operators are omitted from restricted execution environments
  unsafe Open Save
*/

  { "open",       "V:S",      "str",                OpenCmd_S          },
  { "save",       "I:VS",     "vw str",             SaveCmd_VS         },
#define UNSAFE 2
  { "append",     "V:VV",     "vw vw",              AppendCmd_VV       },
  { "at",         "U:VII",    "vw int int",         AtCmd_VII          },
  { "blocked",    "V:V",      "vw",                 BlockedCmd_V       },
  { "clone",      "V:V",      "vw",                 CloneCmd_V         },
  { "coerce",     "U:CS",     "col str",            CoerceCmd_CS       },
  { "colconv",    "C:C",      "col",                ColConvCmd_C       },
  { "colmap",     "V:Vn",     "vw n*",              ColMapCmd_Vn       },
  { "colomit",    "V:Vn",     "vw n*",              ColOmitCmd_Vn      },
  { "compare",    "I:VV",     "vw vw",              CompareCmd_VV      },
  { "compat",     "I:VV",     "vw vw",              CompatCmd_VV       },
  { "concat",     "V:VV",     "vw vw",              ConcatCmd_VV       },
  { "counts",     "C:VN",     "vw N",               CountsCmd_VN       },
  { "countscol",  "C:C",      "col",                CountsColCmd_C     },
  { "countview",  "V:I",      "int",                CountViewCmd_I     },
  { "data",       "V:VX",     "vw ...",             DataCmd_VX         },
  { "def",        "V:OO",     "any any",            DefCmd_OO          },
  { "delete",     "V:VII",    "vw int int",         DeleteCmd_VII      },
  { "emit",       "O:V",      "vw",                 EmitCmd_V          },
  { "except",     "V:VV",     "vw vw",              ExceptCmd_VV       },
  { "first",      "V:VI",     "vw int",             FirstCmd_VI        },
  { "get",        "U:VX",     "vw ...",             GetCmd_VX          },
  { "getcol",     "C:VN",     "vw N",               GetColCmd_VN       },
  { "getinfo",    "C:VVI",    "vw vw int",          GetInfoCmd_VVI     },
  { "group",      "V:VnS",    "vw n* str",          GroupCmd_VnS       },
  { "grouped",    "V:ViiS",   "vw i* i* str",       GroupedCmd_ViiS    },
  { "groupinfo",  "C:V",      "vw",                 GroupInfoCmd_V     },
  { "hashcol",    "C:SC",     "str col",            HashColCmd_SC      },
  { "hashfind",   "I:VIViii", "vw int vw i* i* i*", HashFindCmd_VIViii },
  { "hashinfo",   "C:V",      "vw",                 HashInfoCmd_V      },
  { "hashview",   "C:V",      "vw",                 HashViewCmd_V      },
  { "ijoin",      "V:VV",     "vw vw",              IjoinCmd_VV        },
  { "insert",     "V:VIV",    "vw int vw",          InsertCmd_VIV      },
  { "intersect",  "V:VV",     "vw vw",              IntersectCmd_VV    },
  { "iota",       "C:I",      "int",                IotaCmd_I          },
  { "isectmap",   "C:VV",     "vw vw",              IsectMapCmd_VV     },
  { "join",       "V:VVS",    "vw vw str",          JoinCmd_VVS        },
  { "joininfo",   "C:VV",     "vw vw",              JoinInfoCmd_VV     },
  { "last",       "V:VI",     "vw int",             LastCmd_VI         },
  { "load",       "V:O",      "any",                LoadCmd_O          },
  { "loop",       "O:X",      "...",                LoopCmd_X          },
  { "max",        "O:VNO",    "vw N any",           MaxCmd_VNO         },
  { "mdef",       "V:O",      "any",                MdefCmd_O          },
  { "meta",       "V:V",      "vw",                 MetaCmd_V          },
  { "min",        "O:VNO",    "vw N any",           MinCmd_VNO         },
  { "namecol",    "V:V",      "vw",                 NameColCmd_V       },
  { "names",      "C:V",      "vw",                 NamesCmd_V         },
  { "omitmap",    "C:iI",     "i* int",             OmitMapCmd_iI      },
  { "pair",       "V:VV",     "vw vw",              PairCmd_VV         },
  { "product",    "V:VV",     "vw vw",              ProductCmd_VV      },
  { "ref",        "O:OX",     "any ...",            RefCmd_OX          },
  { "refs",       "I:O",      "any",                RefsCmd_O          },
  { "remap",      "V:Vi",     "vw i*",              RemapCmd_Vi        },
  { "remapsub",   "V:ViII",   "vw i* int int",      RemapSubCmd_ViII   },
  { "rename",     "V:VO",     "vw any",             RenameCmd_VO       },
  { "repeat",     "V:VI",     "vw int",             RepeatCmd_VI       },
  { "replace",    "V:VIIV",   "vw int int vw",      ReplaceCmd_VIIV    },
  { "reverse",    "V:V",      "vw",                 ReverseCmd_V       },
  { "roweq",      "I:IVIV",   "int vw int vw",      RowEqCmd_IVIV      },
  { "rowhash",    "I:VI",     "vw int",             RowHashCmd_VI      },
  { "set",        "V:VIX",    "vw int ...",         SetCmd_VIX         },
  { "size",       "I:V",      "vw",                 SizeCmd_V          },
  { "slice",      "V:VIII",   "vw int int int",     SliceCmd_VIII      },
  { "sort",       "V:V",      "vw",                 SortCmd_V          },
  { "sortmap",    "C:V",      "vw",                 SortMapCmd_V       },
  { "spread",     "V:VI",     "vw int",             SpreadCmd_VI       },
  { "step",       "V:VIIII",  "vw int int int int", StepCmd_VIIII      },
  { "strlookup",  "I:Ss",     "str s*",             StrLookupCmd_Ss    },
  { "structdesc", "S:V",      "vw",                 StructDescCmd_V    },
  { "structure",  "S:V",      "vw",                 StructureCmd_V     },
  { "sum",        "O:VN",     "vw N",               SumCmd_VN          },
  { "tag",        "V:VS",     "vw str",             TagCmd_VS          },
  { "take",       "V:VI",     "vw int",             TakeCmd_VI         },
  { "to",         "O:OO",     "any any",            ToCmd_OO           },
  { "type",       "O:O",      "any",                TypeCmd_O          },
  { "types",      "C:V",      "vw",                 TypesCmd_V         },
  { "ungroup",    "V:VN",     "vw N",               UngroupCmd_VN      },
  { "union",      "V:VV",     "vw vw",              UnionCmd_VV        },
  { "uniqmap",    "C:V",      "vw",                 UniqMapCmd_V       },
  { "unique",     "V:V",      "vw",                 UniqueCmd_V        },
  { "view",       "O:X",      "...",                ViewCmd_X          },
  { "viewconv",   "V:V",      "vw",                 ViewConvCmd_V      },
  { "width",      "I:V",      "vw",                 WidthCmd_V         },
  { "write",      "I:VS",     "vw str",             WriteCmd_VS        },

/* end of generated code */
  { NULL, NULL, NULL, NULL, }
};

#ifndef PACKAGE_VERSION
#define PACKAGE_VERSION "4"
#endif

DLLEXPORT int Vlerq_Init (Tcl_Interp *interp) {
  if (!MyInitStubs(interp) || Tcl_PkgRequire(interp, "Tcl", "8.4", 0) == NULL)
    return TCL_ERROR;

  Tcl_CreateObjCommand(interp, "vlerq", VlerqObjCmd, f_commands, NULL);
  InitShared();
  return Tcl_PkgProvide(interp, "vlerq", PACKAGE_VERSION);
}

DLLEXPORT int Vlerq_SafeInit (Tcl_Interp *interp) {
  if (!MyInitStubs(interp) || Tcl_PkgRequire(interp, "Tcl", "8.4", 0) == NULL)
    return TCL_ERROR;

  /* UNSAFE is defined in the "opdefs_gen.h" file included above */
  Tcl_CreateObjCommand(interp, "vlerq", VlerqObjCmd, f_commands + UNSAFE, NULL);
  InitShared();
  return Tcl_PkgProvide(interp, "vlerq", PACKAGE_VERSION);
}

/* End of generated source code. */
